# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Kaixu Yang
"""Changepoint detection via adaptive lasso."""

import warnings
from datetime import timedelta
from typing import List
from typing import Optional

import numpy as np
import pandas as pd
from pandas.tseries.frequencies import to_offset
from plotly.offline import iplot
from sklearn.base import RegressorMixin
from sklearn.exceptions import ConvergenceWarning
from sklearn.linear_model import LassoCV
from sklearn.linear_model import LinearRegression
from sklearn.linear_model import RidgeCV

from greykite.algo.changepoint.adalasso.changepoints_utils import build_seasonality_feature_df_with_changes
from greykite.algo.changepoint.adalasso.changepoints_utils import build_trend_feature_df_with_changes
from greykite.algo.changepoint.adalasso.changepoints_utils import check_freq_unit_at_most_day
from greykite.algo.changepoint.adalasso.changepoints_utils import combine_detected_and_custom_trend_changepoints
from greykite.algo.changepoint.adalasso.changepoints_utils import compute_fitted_components
from greykite.algo.changepoint.adalasso.changepoints_utils import compute_min_changepoint_index_distance
from greykite.algo.changepoint.adalasso.changepoints_utils import estimate_seasonality_with_detected_changepoints
from greykite.algo.changepoint.adalasso.changepoints_utils import estimate_trend_with_detected_changepoints
from greykite.algo.changepoint.adalasso.changepoints_utils import get_changepoint_dates_from_changepoints_dict
from greykite.algo.changepoint.adalasso.changepoints_utils import get_seasonality_changes_from_adaptive_lasso
from greykite.algo.changepoint.adalasso.changepoints_utils import get_trend_changes_from_adaptive_lasso
from greykite.algo.changepoint.adalasso.changepoints_utils import get_yearly_seasonality_changepoint_dates_from_freq
from greykite.algo.changepoint.adalasso.changepoints_utils import plot_change
from greykite.common.features.timeseries_features import get_evenly_spaced_changepoints_dates
from greykite.common.logging import LoggingLevelEnum
from greykite.common.logging import log_message
from greykite.common.logging import pprint
from greykite.common.python_utils import ignore_warnings


class ChangepointDetector:
    """A class to implement change point detection.

    Currently supports long-term change point detection only. Input is a dataframe with time_col
    indicating the column of time info (the format should be able to be parsed by pd.to_datetime),
    and value_col indicating the column of observed time series values.

    Attributes
    ----------
    original_df : `pandas.DataFrame`
        The original data df, used to retrieve original observations, if aggregation is used in
        fitting change points.
    time_col : `str`
        The column name for time column.
    value_col : `str`
        The column name for value column.
    trend_potential_changepoint_n: `int`
        The number of change points that are evenly distributed over the time period.
    yearly_seasonality_order : `int`
        The yearly seasonality order used when fitting trend.
    y : `pandas.Series`
        The observations after aggregation.
    trend_df : `pandas.DataFrame`
        The augmented df of the original_df, including regressors of trend change points and
        Fourier series for yearly seasonality.
    trend_model : `sklearn.base.RegressionMixin`
        The fitted trend model.
    trend_coef : `numpy.array`
        The estimated trend coefficients.
    trend_intercept : `float`
        The estimated trend intercept.
    adaptive_lasso_coef : `list`
        The list of length two, first element is estimated trend coefficients, and second element
        is intercept, both estimated by adaptive lasso.
    trend_changepoints : `list`
        The list of detected trend change points, parsable by pd.to_datetime
    trend_estimation : `pd.Series`
        The estimated trend with detected trend change points.
    seasonality_df : `pandas.DataFrame`
        The augmented df of ``original_df``, including regressors of seasonality change points with
        different Fourier series frequencies.
    seasonality_changepoints : `dict`
        The dictionary of detected seasonality change points for each component.
        Keys are component names, and values are list of change points.
    seasonality_estimation : `pandas.Series`
        The estimated seasonality with detected seasonality change points.
        The series has the same length as ``original_df``. Index is timestamp, and values
        are the estimated seasonality at each timestamp.
        The seasonality estimation is the estimated of seasonality effect with trend estimated
        by `~greykite.algo.changepoint.adalasso.changepoints_utils.estimate_trend_with_detected_changepoints`
        removed.

    Methods
    -------
    find_trend_changepoints : callable
        Finds the potential trend change points for a given time series df.
    plot : callable
        Plot the results after implementing find_trend_changepoints.
    """

    def __init__(self):
        self.original_df: Optional[pd.DataFrame] = None
        self.time_col: Optional[str] = None
        self.value_col: Optional[str] = None
        self.trend_potential_changepoint_n: Optional[int] = None
        self.yearly_seasonality_order: Optional[int] = None
        self.y: Optional[pd.Series, pd.DataFrame] = None
        self.trend_df: Optional[pd.Series, pd.DataFrame] = None
        self.trend_model: Optional[RegressorMixin] = None
        self.trend_coef: Optional[np.ndarray] = None
        self.trend_intercept: Optional[float] = None
        self.adaptive_lasso_coef: Optional[List] = None
        self.trend_changepoints: Optional[List] = None
        self.trend_estimation: Optional[pd.Series] = None
        self.seasonality_df: Optional[pd.DataFrame] = None
        self.seasonality_changepoints: Optional[dict] = None
        self.seasonality_estimation: Optional[pd.Series] = None

    @ignore_warnings(category=ConvergenceWarning)
    def find_trend_changepoints(
            self,
            df,
            time_col,
            value_col,
            yearly_seasonality_order=8,
            yearly_seasonality_change_freq=None,
            resample_freq="D",
            trend_estimator="ridge",
            adaptive_lasso_initial_estimator="ridge",
            regularization_strength=None,
            actual_changepoint_min_distance="30D",
            potential_changepoint_distance=None,
            potential_changepoint_n=100,
            no_changepoint_distance_from_begin=None,
            no_changepoint_proportion_from_begin=0.0,
            no_changepoint_distance_from_end=None,
            no_changepoint_proportion_from_end=0.0):
        """Finds trend change points automatically by adaptive lasso.

        The algorithm does an aggregation with a user-defined frequency, defaults daily.

        If ``potential_changepoint_distance`` is not given,  ``potential_changepoint_n``
        potential change points are evenly distributed over the time period, else
        ``potential_changepoint_n`` is overridden by::

                total_time_length / ``potential_changepoint_distance``

        Users can specify either ``no_changepoint_proportion_from_end`` to specify what proportion
        from the end of data they do not want changepoints, or ``no_changepoint_distance_from_end``
        (overrides ``no_changepoint_proportion_from_end``) to specify how long from the end they
        do not want change points.

        Then all potential change points will be selected by adaptive lasso, with the initial
        estimator specified by ``adaptive_lasso_initial_estimator``. If user specifies
        ``regularization_strength``, then the adaptive lasso will be run with a single tuning
        parameter calculated based on user provided prior, else a cross-validation will be run to
        automatically select the tuning parameter.

        A yearly seasonality is also fitted at the same time, preventing trend from catching
        yearly periodical changes.

        A rule-based guard function is applied at the end to ensure change points are not
        too close, as specified by ``actual_changepoint_min_distance``.

        Parameters
        ----------
        df: `pandas.DataFrame`
            The data df
        time_col : `str`
            Time column name in ``df``
        value_col : `str`
            Value column name in ``df``
        yearly_seasonality_order : `int`, default 8
            Fourier series order to capture yearly seasonality.
        yearly_seasonality_change_freq : `DateOffset`, `Timedelta` or `str` or `None`, default `None`
            How often to change the yearly seasonality model. Set to `None` to disable this feature.

            This is useful if you have more than 2.5 years of data and the detected trend without this
            feature is inaccurate because yearly seasonality changes over the training period.
            Modeling yearly seasonality separately over the each period can prevent trend changepoints
            from fitting changes in yearly seasonality. For example, if you have 2.5 years of data and
            yearly seasonality increases in magnitude after the first year, setting this parameter to
            "365D" will model each year's yearly seasonality differently and capture both shapes.
            However, without this feature, both years will have the same yearly seasonality, roughly
            the average effect across the training set.

            Note that if you use `str` as input, the maximal supported unit is day, i.e.,
            you might use "200D" but not "12M" or "1Y".
        resample_freq : `DateOffset`, `Timedelta` or `str`, default "D".
            The frequency to aggregate data.
            Coarser aggregation leads to fitting longer term trends.
        trend_estimator : `str` in ["ridge", "lasso" or "ols"], default "ridge".
            The estimator to estimate trend. The estimated trend is only for plotting purposes.
            'ols' is not recommended when ``yearly_seasonality_order`` is specified other than 0,
            because significant over-fitting will happen.
            In this case, the given value is overridden by "ridge".
        adaptive_lasso_initial_estimator : `str` in ["ridge", "lasso" or "ols"], default "ridge".
            The initial estimator to compute adaptive lasso weights
        regularization_strength : `float` in [0, 1] or `None`
            The regularization for change points. Greater value implies fewer change points.
            0 indicates all change points, and 1 indicates no change point.
            If `None`, the turning parameter will be selected by cross-validation.
            If a value is given, it will be used as the tuning parameter.
        actual_changepoint_min_distance : `DateOffset`, `Timedelta` or `str`, default "30D"
            The minimal distance allowed between detected change points. If consecutive change points
            are within this minimal distance, the one with smaller absolute change coefficient will
            be dropped.
            Note: maximal unit is 'D', i.e., you may use units no more than 'D' such as
            '10D', '5H', '100T', '200S'. The reason is that 'W', 'M' or higher has either
            cycles or indefinite number of days, thus is not parsable by pandas as timedelta.
        potential_changepoint_distance : `DateOffset`, `Timedelta`, `str` or None, default None
            The distance between potential change points.
            If provided, will override the parameter ``potential_changepoint_n``.
            Note: maximal unit is 'D', i.e., you may only use units no more than 'D' such as
            '10D', '5H', '100T', '200S'. The reason is that 'W', 'M' or higher has either
            cycles or indefinite number of days, thus is not parsable by pandas as timedelta.
        potential_changepoint_n : `int`, default 100
            Number of change points to be evenly distributed, recommended 1-2 per month, based
            on the training data length.
        no_changepoint_distance_from_begin : `DateOffset`, `Timedelta`, `str` or None, default None
            The length of time from the beginning of training data, within which no change point will be placed.
            If provided, will override the parameter ``no_changepoint_proportion_from_begin``.
            Note: maximal unit is 'D', i.e., you may only use units no more than 'D' such as
            '10D', '5H', '100T', '200S'. The reason is that 'W', 'M' or higher has either
            cycles or indefinite number of days, thus is not parsable by pandas as timedelta.
        no_changepoint_proportion_from_begin : `float` in [0, 1], default 0.0.
            ``potential_changepoint_n`` change points will be placed evenly over the whole training period,
            however, change points that are located within the first ``no_changepoint_proportion_from_begin``
            proportion of training period will not be used for change point detection.
        no_changepoint_distance_from_end : `DateOffset`, `Timedelta`, `str` or None, default None
            The length of time from the end of training data, within which no change point will be placed.
            If provided, will override the parameter ``no_changepoint_proportion_from_end``.
            Note: maximal unit is 'D', i.e., you may only use units no more than 'D' such as
            '10D', '5H', '100T', '200S'. The reason is that 'W', 'M' or higher has either
            cycles or indefinite number of days, thus is not parsable by pandas as timedelta.
        no_changepoint_proportion_from_end : `float` in [0, 1], default 0.0.
            ``potential_changepoint_n`` change points will be placed evenly over the whole training period,
            however, change points that are located within the last ``no_changepoint_proportion_from_end``
            proportion of training period will not be used for change point detection.

        Return
        ------
        result : `dict`
            result dictionary with keys:

            ``"trend_feature_df"`` : `pandas.DataFrame`
                The augmented df for change detection, in other words, the design matrix for
                the regression model. Columns:

                    - 'changepoint0': regressor for change point 0, equals the continuous time
                      of the observation minus the continuous time for time of origin.
                    - ...
                    - 'changepoint{potential_changepoint_n}': regressor for change point
                      {potential_changepoint_n}, equals the continuous time of the observation
                      minus the continuous time of the {potential_changepoint_n}th change point.
                    - 'cos1_conti_year_yearly': cosine yearly seasonality regressor of first order.
                    - 'sin1_conti_year_yearly': sine yearly seasonality regressor of first order.
                    - ...
                    - 'cos{yearly_seasonality_order}_conti_year_yearly' : cosine yearly seasonality
                      regressor of {yearly_seasonality_order}th order.
                    - 'sin{yearly_seasonality_order}_conti_year_yearly' : sine yearly seasonality
                      regressor of {yearly_seasonality_order}th order.

            ``"trend_changepoints"`` : `list`
                The list of detected change points.
            ``"changepoints_dict"`` : `dict`
                The change point dictionary that is compatible as an input with
                `~greykite.algo.forecast.silverkite.forecast_silverkite.SilverkiteForecast.forecast`
            ``"trend_estimation"`` : `pandas.Series`
                The estimated trend with detected trend change points.
        """
        # Checks parameter rationality
        if potential_changepoint_n < 0:
            raise ValueError("potential_changepoint_n can not be negative. "
                             "A large number such as 100 is recommended")
        if yearly_seasonality_order < 0:
            raise ValueError("year_seasonality_order can not be negative. "
                             "A number less than or equal to 10 is recommended")
        if df.dropna().shape[0] < 5:
            raise ValueError("Change point detector does not work for less than "
                             "5 observations. Please increase sample size.")
        if no_changepoint_proportion_from_begin < 0 or no_changepoint_proportion_from_begin > 1:
            raise ValueError("no_changepoint_proportion_from_begin needs to be between 0 and 1.")
        if no_changepoint_proportion_from_end < 0 or no_changepoint_proportion_from_end > 1:
            raise ValueError("no_changepoint_proportion_from_end needs to be between 0 and 1.")
        if no_changepoint_distance_from_begin is not None:
            check_freq_unit_at_most_day(no_changepoint_distance_from_begin, "no_changepoint_distance_from_begin")
            data_length = pd.to_datetime(df[time_col].iloc[-1]) - pd.to_datetime(df[time_col].iloc[0])
            no_changepoint_proportion_from_begin = to_offset(no_changepoint_distance_from_begin).delta / data_length
            no_changepoint_proportion_from_begin = min(no_changepoint_proportion_from_begin, 1)
        if no_changepoint_distance_from_end is not None:
            check_freq_unit_at_most_day(no_changepoint_distance_from_end, "no_changepoint_distance_from_end")
            data_length = pd.to_datetime(df[time_col].iloc[-1]) - pd.to_datetime(df[time_col].iloc[0])
            no_changepoint_proportion_from_end = to_offset(no_changepoint_distance_from_end).delta / data_length
            no_changepoint_proportion_from_end = min(no_changepoint_proportion_from_end, 1)
        if potential_changepoint_distance is not None:
            check_freq_unit_at_most_day(potential_changepoint_distance, "potential_changepoint_distance")
            data_length = pd.to_datetime(df[time_col].iloc[-1]) - pd.to_datetime(df[time_col].iloc[0])
            potential_changepoint_n = data_length // to_offset(potential_changepoint_distance).delta
        if regularization_strength is not None and (regularization_strength < 0 or regularization_strength > 1):
            raise ValueError("regularization_strength must be between 0.0 and 1.0.")
        self.trend_potential_changepoint_n = potential_changepoint_n
        self.time_col = time_col
        self.value_col = value_col
        self.original_df = df.copy()
        # Resamples df to get a coarser granularity to get rid of shorter seasonality.
        # The try except below speeds up unnecessary datetime transformation.
        try:
            df_resample = df.resample(resample_freq, on=time_col).mean().reset_index()
        except TypeError:
            df = df.copy()
            df[time_col] = pd.to_datetime(df[time_col])
            df_resample = df.resample(resample_freq, on=time_col).mean().reset_index()
        # The ``df.resample`` function creates NA when the original df has a missing observation
        # or its value is NA.
        # The estimation algorithm does not allow NA, so we drop those rows.
        df_resample = df_resample.dropna()
        self.original_df[time_col] = df[time_col]
        # Prepares response df.
        y = df_resample[value_col]
        y.index = df_resample[time_col]
        self.y = y
        # Prepares trend feature df.
        # Potential changepoints are placed uniformly among rows without a missing value, after resampling.
        trend_df = build_trend_feature_df_with_changes(
            df=df_resample,
            time_col=time_col,
            changepoints_dict={
                "method": "uniform",
                "n_changepoints": potential_changepoint_n
            }
        )
        # Gets changepoint features only in range filtered by ``no_changepoint_proportion_from_begin`` and
        # ``no_changepoint_proportion_from_end`` of time period.
        n_changepoints_within_range_begin = int(potential_changepoint_n * no_changepoint_proportion_from_begin)
        n_changepoints_within_range_end = int(potential_changepoint_n * (1 - no_changepoint_proportion_from_end))
        if n_changepoints_within_range_begin < n_changepoints_within_range_end:
            trend_df = trend_df.iloc[:, [0] + list(range(n_changepoints_within_range_begin + 1, n_changepoints_within_range_end + 1))]
        else:
            # Linear growth term only.
            trend_df = trend_df.iloc[:, [0]]
        # Builds yearly seasonality feature df
        if yearly_seasonality_order is not None and yearly_seasonality_order > 0:
            self.yearly_seasonality_order = yearly_seasonality_order
            # Gets yearly seasonality changepoints, allowing varying yearly seasonality coefficients
            # to capture yearly seasonality shape change.
            yearly_seasonality_changepoint_dates = get_yearly_seasonality_changepoint_dates_from_freq(
                df=df,
                time_col=time_col,
                yearly_seasonality_change_freq=yearly_seasonality_change_freq)
            long_seasonality_df = build_seasonality_feature_df_with_changes(
                df=df_resample,
                time_col=time_col,
                changepoints_dict=dict(
                    method="custom",
                    dates=yearly_seasonality_changepoint_dates),
                fs_components_df=pd.DataFrame({
                    "name": ["conti_year"],
                    "period": [1.0],
                    "order": [yearly_seasonality_order],
                    "seas_names": ["yearly"]})
            )
            trend_df = pd.concat([trend_df, long_seasonality_df], axis=1)
        trend_df.index = df_resample[time_col]
        self.trend_df = trend_df
        # Estimates trend.
        if trend_estimator not in ["ridge", "lasso", "ols"]:
            warnings.warn("trend_estimator not in ['ridge', 'lasso', 'ols'], "
                          "estimating using ridge")
            trend_estimator = 'ridge'
        if trend_estimator == 'ols' and yearly_seasonality_order > 0:
            warnings.warn("trend_estimator = 'ols' with year_seasonality_order > 0 may create "
                          "over-fitting, trend_estimator has been set to 'ridge'.")
            trend_estimator = 'ridge'
        fit_algorithm_dict = {
            "ridge": RidgeCV,
            "lasso": LassoCV,
            "ols": LinearRegression
        }
        trend_model = fit_algorithm_dict[trend_estimator]().fit(trend_df.values, y.values)
        self.trend_model = trend_model
        self.trend_coef, self.trend_intercept = trend_model.coef_.ravel(), trend_model.intercept_.ravel()
        # Fetches change point dates for reference as datetime format.
        changepoint_dates = get_evenly_spaced_changepoints_dates(
            df=df_resample,
            time_col=time_col,
            n_changepoints=potential_changepoint_n
        )
        # Gets the changepoint dates filtered by ``no_changepoint_proportion_from_begin`` and ``no_changepoint_proportion_from_end``.
        if n_changepoints_within_range_begin < n_changepoints_within_range_end:
            changepoint_dates = changepoint_dates.iloc[[0] + list(range(n_changepoints_within_range_begin + 1, n_changepoints_within_range_end + 1))]
        else:
            # Linear growth term only.
            changepoint_dates = changepoint_dates.iloc[[0]]
        # Calculates the minimal allowed change point index distance.
        min_changepoint_index_distance = compute_min_changepoint_index_distance(
            df=df_resample,
            time_col=time_col,
            n_changepoints=potential_changepoint_n,
            min_distance_between_changepoints=actual_changepoint_min_distance
        )
        # Uses adaptive lasso to select change points.
        if adaptive_lasso_initial_estimator not in ['ridge', 'lasso', 'ols']:
            warnings.warn("adaptive_lasso_initial_estimator not in ['ridge', 'lasso', 'ols'], "
                          "estimating with ridge")
            adaptive_lasso_initial_estimator = "ridge"
        if adaptive_lasso_initial_estimator == trend_estimator:
            # When ``adaptive_lasso_initial_estimator`` is the same as ``trend_estimator``, the
            # estimated trend coefficients will be used to calculate the weights. The
            # ``get_trend_changes_from_adaptive_lasso`` function recognizes ``initial_coef`` as
            # `numpy.array` and calculates the weights directly.
            trend_changepoints, self.adaptive_lasso_coef = get_trend_changes_from_adaptive_lasso(
                x=trend_df.values,
                y=y.values,
                changepoint_dates=changepoint_dates,
                initial_coef=self.trend_coef,
                min_index_distance=min_changepoint_index_distance,
                regularization_strength=regularization_strength
            )
        else:
            # When ``adaptive_lasso_initial_estimator`` is different from ``trend_estimator``, the
            # ``adaptive_lasso_initial_estimator`` as a `str` will be passed. The
            # ``get_trend_changes_from_adaptive_lasso`` function recognizes ``initial_coef`` as
            # `str` and calculates the initial estimator with the corresponding estimator first
            # then calculates the weights.
            trend_changepoints, self.adaptive_lasso_coef = get_trend_changes_from_adaptive_lasso(
                x=trend_df.values,
                y=y.values,
                changepoint_dates=changepoint_dates,
                initial_coef=adaptive_lasso_initial_estimator,
                min_index_distance=min_changepoint_index_distance,
                regularization_strength=regularization_strength
            )
        # Checks if the beginning date is picked as a change point. If yes, drop it, because we
        # always include the growth term in our model.
        trend_changepoints = [cp for cp in trend_changepoints if cp > max(df_resample[time_col][0], df[time_col][0])]
        self.trend_changepoints = trend_changepoints
        # logging
        log_message(f"The detected trend change points are\n{trend_changepoints}", LoggingLevelEnum.INFO)
        # Creates changepoints_dict for silverkite to use.
        changepoints_dict = {
            "method": "custom",
            "dates": trend_changepoints
        }
        # Computes trend estimates for seasonality use.
        trend_estimation = estimate_trend_with_detected_changepoints(
            df=df,
            time_col=time_col,
            value_col=value_col,
            changepoints=trend_changepoints
        )
        self.trend_estimation = trend_estimation
        result = {
            "trend_feature_df": trend_df,
            "trend_changepoints": trend_changepoints,
            "changepoints_dict": changepoints_dict,
            "trend_estimation": trend_estimation
        }
        return result

    @ignore_warnings(category=ConvergenceWarning)
    def find_seasonality_changepoints(
            self,
            df,
            time_col,
            value_col,
            seasonality_components_df=pd.DataFrame({
                "name": ["tod", "tow", "conti_year"],
                "period": [24.0, 7.0, 1.0],
                "order": [3, 3, 5],
                "seas_names": ["daily", "weekly", "yearly"]}),
            resample_freq="H",
            regularization_strength=0.6,
            actual_changepoint_min_distance="30D",
            potential_changepoint_distance=None,
            potential_changepoint_n=50,
            no_changepoint_distance_from_end=None,
            no_changepoint_proportion_from_end=0.0,
            trend_changepoints=None):
        """Finds the seasonality change points (defined as the time points where seasonality
        magnitude changes, i.e., the time series becomes "fatter" or "thinner".)

        Subtracts the estimated trend from the original time series first,
        then uses regression-based regularization methods to select important seasonality
        change points. Regressors are built from truncated Fourier series.

        If you have run ``find_trend_changepoints`` before running ``find_seasonality_changepoints``
        with the same df, the estimated trend will be automatically used for removing trend in
        ``find_seasonality_changepoints``.
        Otherwise, ``find_trend_changepoints`` will be run automatically with the same parameters
        as you passed to ``find_seasonality_changepoints``. If you do not want to use the same
        parameters, run ``find_trend_changepoints`` with your desired parameter before calling
        ``find_seasonality_changepoints``.

        The algorithm does an aggregation with a user-defined frequency, default hourly.

        The regression features consists of ``potential_changepoint_n`` + 1 blocks of
        predictors. The first block consists of Fourier series according to
        ``seasonality_components_df``, and other blocks are a copy of the first block
        truncated at the corresponding potential change point.

        If ``potential_changepoint_distance`` is not given,  ``potential_changepoint_n``
        potential change points are evenly distributed over the time period, else
        ``potential_changepoint_n`` is overridden by::

                total_time_length / ``potential_changepoint_distance``

        Users can specify either ``no_changepoint_proportion_from_end`` to specify what proportion
        from the end of data they do not want changepoints, or ``no_changepoint_distance_from_end``
        (overrides ``no_changepoint_proportion_from_end``) to specify how long from the end they
        do not want change points.

        Then all potential change points will be selected by adaptive lasso, with the initial
        estimator specified by ``adaptive_lasso_initial_estimator``. The regularization strength
        is specified by ``regularization_strength``, which lies between 0 and 1.

        A rule-based guard function is applied at the end to ensure change points are not
        too close, as specified by ``actual_changepoint_min_distance``.

        Parameters
        ----------
        df: `pandas.DataFrame`
            The data df
        time_col : `str`
            Time column name in ``df``
        value_col : `str`
            Value column name in ``df``
        seasonality_components_df : `pandas.DataFrame`
            The df to generate seasonality design matrix, which is compatible with
            ``seasonality_components_df`` in
            `~greykite.algo.changepoint.adalasso.changepoint_detector.ChangepointDetector.find_seasonality_changepoints`
        resample_freq : `DateOffset, Timedelta or str`, default "H".
            The frequency to aggregate data.
            Coarser aggregation leads to fitting longer term trends.
        regularization_strength : `float` in [0, 1] or `None`, default 0.6.
            The regularization for change points. Greater value implies fewer change points.
            0 indicates all change points, and 1 indicates no change point.
            If `None`, the turning parameter will be selected by cross-validation.
            If a value is given, it will be used as the tuning parameter.
            Here "None" is not recommended, because seasonality change has different levels,
            and automatic selection by cross-validation may produce more change points than
            desired. Practically, 0.6 is a good choice for most cases. Tuning around
            0.6 is recommended.
        actual_changepoint_min_distance : `DateOffset`, `Timedelta` or `str`, default "30D"
            The minimal distance allowed between detected change points. If consecutive change points
            are within this minimal distance, the one with smaller absolute change coefficient will
            be dropped.
            Note: maximal unit is 'D', i.e., you may use units no more than 'D' such as
            '10D', '5H', '100T', '200S'. The reason is that 'W', 'M' or higher has either
            cycles or indefinite number of days, thus is not parsable by pandas as timedelta.
        potential_changepoint_distance : `DateOffset`, `Timedelta`, `str` or None, default None
            The distance between potential change points.
            If provided, will override the parameter ``potential_changepoint_n``.
            Note: maximal unit is 'D', i.e., you may only use units no more than 'D' such as
            '10D', '5H', '100T', '200S'. The reason is that 'W', 'M' or higher has either
            cycles or indefinite number of days, thus is not parsable by pandas as timedelta.
        potential_changepoint_n : `int`, default 50
            Number of change points to be evenly distributed, recommended 1 per month, based
            on the training data length.
        no_changepoint_distance_from_end : `DateOffset`, `Timedelta`, `str` or None, default None
            The length of time from the end of training data, within which no change point will be placed.
            If provided, will override the parameter ``no_changepoint_proportion_from_end``.
            Note: maximal unit is 'D', i.e., you may only use units no more than 'D' such as
            '10D', '5H', '100T', '200S'. The reason is that 'W', 'M' or higher has either
            cycles or indefinite number of days, thus is not parsable by pandas as timedelta.
        no_changepoint_proportion_from_end : `float` in [0, 1], default 0.0.
            ``potential_changepoint_n`` change points will be placed evenly over the whole training period,
            however, only change points that are not located within the last ``no_changepoint_proportion_from_end``
            proportion of training period will be used for change point detection.
        trend_changepoints : `list` or None
            A list of user specified trend change points, used to estimated the trend to be removed
            from the time series before detecting seasonality change points. If provided, the algorithm
            will not check existence of detected trend change points or run ``find_trend_changepoints``,
            but will use these change points directly for trend estimation.

        Return
        ------
        result : `dict`
            result dictionary with keys:

            ``"seasonality_feature_df"`` : `pandas.DataFrame`
                The augmented df for seasonality changepoint detection, in other words, the design matrix for
                the regression model. Columns:

                    - "cos1_tod_daily": cosine daily seasonality regressor of first order at change point 0.
                    - "sin1_tod_daily": sine daily seasonality regressor of first order at change point 0.
                    - ...
                    - "cos1_conti_year_yearly": cosine yearly seasonality regressor of first order at
                      change point 0.
                    - "sin1_conti_year_yearly": sine yearly seasonality regressor of first order at
                      change point 0.
                    - ...
                    - "cos{daily_seasonality_order}_tod_daily_cp{potential_changepoint_n}" : cosine
                      daily seasonality regressor of {yearly_seasonality_order}th order at change point
                      {potential_changepoint_n}.
                    - "sin{daily_seasonality_order}_tod_daily_cp{potential_changepoint_n}" : sine
                      daily seasonality regressor of {yearly_seasonality_order}th order at change point
                      {potential_changepoint_n}.
                    - ...
                    - "cos{yearly_seasonality_order}_conti_year_yearly_cp{potential_changepoint_n}" : cosine
                      yearly seasonality regressor of {yearly_seasonality_order}th order at change point
                      {potential_changepoint_n}.
                    - "sin{yearly_seasonality_order}_conti_year_yearly_cp{potential_changepoint_n}" : sine
                      yearly seasonality regressor of {yearly_seasonality_order}th order at change point
                      {potential_changepoint_n}.

            ``"seasonality_changepoints"`` : `dict`[`list`[`datetime`]]
                The dictionary of detected seasonality change points for each component.
                Keys are component names, and values are list of change points.
            ``"seasonality_estimation"`` : `pandas.Series`
                The estimated seasonality with detected seasonality change points.
                    The series has the same length as ``original_df``. Index is timestamp, and values
                    are the estimated seasonality at each timestamp.
                    The seasonality estimation is the estimated of seasonality effect with trend estimated
                    by `~greykite.algo.changepoint.adalasso.changepoints_utils.estimate_trend_with_detected_changepoints`
                    removed.
            ``"seasonality_components_df`` : `pandas.DataFrame`
                The processed ``seasonality_components_df``. Daily component row is removed if
                inferred frequency or aggregation frequency is at least one day.
        """
        # Checks parameter rationality.
        if potential_changepoint_n < 0:
            raise ValueError("potential_changepoint_n can not be negative. "
                             "A large number such as 50 is recommended")
        if df.dropna().shape[0] < 5:
            raise ValueError("Change point detector does not work for less than "
                             "5 observations. Please increase sample size.")
        if no_changepoint_proportion_from_end < 0 or no_changepoint_proportion_from_end > 1:
            raise ValueError("``no_changepoint_proportion_from_end`` needs to be between 0 and 1.")
        if no_changepoint_distance_from_end is not None:
            check_freq_unit_at_most_day(no_changepoint_distance_from_end, "no_changepoint_distance_from_end")
            data_length = pd.to_datetime(df[time_col].iloc[-1]) - pd.to_datetime(df[time_col].iloc[0])
            no_changepoint_proportion_from_end = to_offset(no_changepoint_distance_from_end).delta / data_length
        if potential_changepoint_distance is not None:
            check_freq_unit_at_most_day(potential_changepoint_distance, "potential_changepoint_distance")
            data_length = pd.to_datetime(df[time_col].iloc[-1]) - pd.to_datetime(df[time_col].iloc[0])
            potential_changepoint_n = data_length // to_offset(potential_changepoint_distance).delta
        if regularization_strength is None:
            warnings.warn("regularization_strength is set to None. This will trigger cross-validation to "
                          "select the tuning parameter which might result in too many change points. "
                          "Keep the default value or tuning around it is recommended.")
        if regularization_strength is not None and (regularization_strength < 0 or regularization_strength > 1):
            raise ValueError("regularization_strength must be between 0.0 and 1.0.")
        df[time_col] = pd.to_datetime(df[time_col])
        # If user provides a list of trend change points, these points will be used to estimate trend.
        if trend_changepoints is not None:
            trend_estimation = estimate_trend_with_detected_changepoints(
                df=df,
                time_col=time_col,
                value_col=value_col,
                changepoints=trend_changepoints
            )
            self.trend_changepoints = trend_changepoints
            self.trend_estimation = trend_estimation
            self.original_df = df
            self.time_col = time_col
            self.value_col = value_col
            self.y = df[value_col]
            self.y.index = df[time_col]
        # If user doesn't provide trend change points, the trend change points will be found automatically.
        else:
            # Checks if trend change point is available.
            # Runs trend change point detection with default value if not.
            compare_df = df.copy()
            if time_col != self.time_col or value_col != self.value_col:
                compare_df.rename({time_col: self.time_col, value_col: self.value_col}, axis=1, inplace=True)
            if (self.original_df is not None
                    and self.original_df[[self.time_col, self.value_col]].equals(
                        compare_df[[self.time_col, self.value_col]])
                    and self.trend_estimation is not None):
                # If the passed df is the same as ``self.original_df``, then the previous
                # ``self.trend_estimation`` is to be subtracted from the time series.
                trend_estimation = self.trend_estimation
                warnings.warn("Trend changepoints are already identified, using past trend estimation. "
                              "If you would like to run trend change point detection again, "
                              "please call ``find_trend_changepoints`` with desired parameters "
                              "before calling ``find_seasonality_changepoints``.")
            else:
                # If the passed df is different from ``self.original_df``, then trend change point
                # detection algorithm is run first, and trend estimation is calculated afterward.
                # In this case, the parameters passed to ``find_seasonality_changepoints`` are also
                # passed to ``find_trend_changepoint``.
                # If you do not want the parameters passed, run ``find_trend_changepoints`` with
                # desired parameters before calling ``find_seasonality_changepoints``.
                trend_result = self.find_trend_changepoints(
                    df=df,
                    time_col=time_col,
                    value_col=value_col,
                    actual_changepoint_min_distance=actual_changepoint_min_distance,
                    no_changepoint_distance_from_end=no_changepoint_distance_from_end,
                    no_changepoint_proportion_from_end=no_changepoint_proportion_from_end
                )
                warnings.warn(f"Trend changepoints are not identified for the input dataframe, "
                              f"triggering trend change point detection with parameters"
                              f"actual_changepoint_min_distance={actual_changepoint_min_distance}\n"
                              f"no_changepoint_proportion_from_end={no_changepoint_proportion_from_end}\n"
                              f"no_changepoint_distance_from_end={no_changepoint_distance_from_end}\n"
                              f" Found trend change points\n{self.trend_changepoints}\n"
                              "If you would like to run trend change point detection with customized "
                              "parameters, please call ``find_trend_changepoints`` with desired parameters "
                              "before calling ``find_seasonality_changepoints``.")
                trend_estimation = trend_result["trend_estimation"]
        # Splits trend effects from time series.
        df_without_trend = df.copy()
        df_without_trend[value_col] -= trend_estimation.values
        # Aggregates df.
        df_resample = df_without_trend.resample(resample_freq, on=time_col).mean().reset_index()
        df_resample = df_resample.dropna()
        # Removes daily component from seasonality_components_df if data has minimum freq daily.
        freq_at_least_day = (min(np.diff(df_resample[time_col]).astype("timedelta64[s]")) >= timedelta(days=1))
        if (freq_at_least_day
                and "daily" in seasonality_components_df["seas_names"].tolist()):
            warnings.warn("Inferred minimum data frequency is at least 1 day, daily component is "
                          "removed from seasonality_components_df.")
            seasonality_components_df = seasonality_components_df.loc[
                seasonality_components_df["seas_names"] != "daily"]
        # Builds seasonality feature df.
        seasonality_df = build_seasonality_feature_df_with_changes(
            df=df_resample,
            time_col=time_col,
            fs_components_df=seasonality_components_df,
            changepoints_dict={
                "method": "uniform",
                "n_changepoints": potential_changepoint_n
            }
        )
        # Eliminates change points from the end
        # the generated seasonality_df has {``potential_changepoint_n`` + 1} blocks, where there
        # are {sum_i(order of component i) * 2} columns consisting of the cosine and sine functions
        # for each order for each component.
        # The selection below selects the first {``n_changepoints_within_range`` + 1} columns,
        # which corresponds to the regular block (first block) and the blocks that correspond
        # to the change points that are within range.
        n_changepoints_within_range = int(potential_changepoint_n * (1 - no_changepoint_proportion_from_end))
        orders = seasonality_components_df["order"].tolist()
        seasonality_df = seasonality_df.iloc[:, :(n_changepoints_within_range + 1) * sum(orders) * 2]
        self.seasonality_df = seasonality_df
        # Fetches change point dates for reference as datetime format.
        changepoint_dates = get_evenly_spaced_changepoints_dates(
            df=df,
            time_col=time_col,
            n_changepoints=potential_changepoint_n
        )
        # Gets the changepoint dates that are not within ``no_changepoint_proportion_from_end``.
        changepoint_dates = changepoint_dates.iloc[0: n_changepoints_within_range + 1]
        # Calculates the minimal allowed change point index distance.
        min_changepoint_index_distance = compute_min_changepoint_index_distance(
            df=df_resample,
            time_col=time_col,
            n_changepoints=potential_changepoint_n,
            min_distance_between_changepoints=actual_changepoint_min_distance
        )
        seasonality_changepoints = get_seasonality_changes_from_adaptive_lasso(
            x=seasonality_df.values,
            y=df_resample[value_col].values,
            changepoint_dates=changepoint_dates,
            initial_coef="lasso",
            seasonality_components_df=seasonality_components_df,
            min_index_distance=min_changepoint_index_distance,
            regularization_strength=regularization_strength
        )
        # Checks if the beginning date is picked as a change point. If yes, drop it, because we
        # always include the overall seasonality term in our model.
        for key in seasonality_changepoints.keys():
            if df_resample[time_col][0] in seasonality_changepoints[key]:
                seasonality_changepoints[key] = seasonality_changepoints[key][1:]
        self.seasonality_changepoints = seasonality_changepoints
        # logging
        log_message(f"The detected seasonality changepoints are\n"
                    f"{pprint(seasonality_changepoints)}", LoggingLevelEnum.INFO)
        # Performs a seasonality estimation for plotting purposes.
        seasonality_estimation = estimate_seasonality_with_detected_changepoints(
            df=df_without_trend,
            time_col=time_col,
            value_col=value_col,
            seasonality_changepoints=seasonality_changepoints,
            seasonality_components_df=seasonality_components_df
        )
        self.seasonality_estimation = seasonality_estimation
        result = {
            "seasonality_feature_df": seasonality_df,
            "seasonality_changepoints": seasonality_changepoints,
            "seasonality_estimation": seasonality_estimation,
            "seasonality_components_df": seasonality_components_df
        }
        return result

    def plot(
            self,
            observation=True,
            observation_original=True,
            trend_estimate=True,
            trend_change=True,
            yearly_seasonality_estimate=False,
            adaptive_lasso_estimate=False,
            seasonality_change=False,
            seasonality_change_by_component=True,
            seasonality_estimate=False,
            plot=True):
        """Makes a plot to show the observations/estimations/change points.

        In this function, component parameters specify if each component in the plot is
        included or not. These are `bool` variables.
        For those components that are set to True, their values will be replaced by the
        corresponding data. Other components values will be set to None. Then these variables
        will be fed into
        `~greykite.algo.changepoint.adalasso.changepoints_utils.plot_change`

        Parameters
        ----------
        observation : `bool`
            Whether to include observation
        observation_original : `bool`
            Set True to plot original observations, and False to plot aggregated observations.
            No effect is ``observation`` is False
        trend_estimate : `bool`
            Set True to add trend estimation.
        trend_change : `bool`
            Set True to add change points.
        yearly_seasonality_estimate : `bool`
            Set True to add estimated yearly seasonality.
        adaptive_lasso_estimate : `bool`
            Set True to add adaptive lasso estimated trend.
        seasonality_change : `bool`
            Set True to add seasonality change points.
        seasonality_change_by_component : `bool`
            If true, seasonality changes will be plotted separately for different components,
            else all will be in the same symbol.
            No effect if ``seasonality_change`` is False
        seasonality_estimate : `bool`
            Set True to add estimated seasonality.
            The seasonality if plotted around trend, so the actual seasonality shown is
            trend estimation + seasonality estimation.
        plot : `bool`, default True
            Set to True to display the plot, and set to False to return the plotly figure object.

        Returns
        -------
        None (if ``plot`` == True)
            The function shows a plot.
        fig : `plotly.graph_objs.Figure`
            The plot object.
        """
        # Adds observation
        if observation:
            if observation_original:
                observation = self.original_df[self.value_col]
                observation.index = self.original_df[self.time_col]
            else:
                observation = self.y
        else:
            observation = None
        # Adds trend estimation
        if trend_estimate:
            trend_estimate = compute_fitted_components(
                x=self.trend_df,
                coef=self.trend_coef,
                regex='^changepoint',
                include_intercept=True,
                intercept=self.trend_intercept
            )
        else:
            trend_estimate = None
        # Adds trend change points
        if trend_change:
            if self.trend_changepoints is None:
                warnings.warn("You haven't run trend change point detection algorithm yet. "
                              "Please call find_trend_changepoints first.")
            trend_change = self.trend_changepoints
        else:
            trend_change = None
        # Adds yearly seasonality estimates
        if yearly_seasonality_estimate:
            yearly_seasonality_estimate = compute_fitted_components(
                x=self.trend_df,
                coef=self.trend_coef,
                regex='^.*yearly.*$',
                include_intercept=False
            )
        else:
            yearly_seasonality_estimate = None
        # Adds adaptive lasso trend estimates
        if adaptive_lasso_estimate and self.adaptive_lasso_coef is not None:
            adaptive_lasso_estimate = compute_fitted_components(
                x=self.trend_df,
                coef=self.adaptive_lasso_coef[1],
                regex='^changepoint',
                include_intercept=True,
                intercept=self.adaptive_lasso_coef[0])
        else:
            adaptive_lasso_estimate = None
        # Adds seasonality change points
        if seasonality_change:
            if self.seasonality_changepoints is None:
                warnings.warn("You haven't run seasonality change point detection algorithm yet. "
                              "Please call find_seasonality_changepoints first.")
            if seasonality_change_by_component:
                seasonality_change = self.seasonality_changepoints
            else:
                seasonality_change = []
                for key in self.seasonality_changepoints.keys():
                    seasonality_change += self.seasonality_changepoints[key]
        else:
            seasonality_change = None
        # Adds seasonality estimates
        if seasonality_estimate:
            if self.seasonality_estimation is None:
                warnings.warn("You haven't run seasonality change point detection algorithm yet. "
                              "Please call find_seasonality_changepoints first.")
                seasonality_estimate = None
            else:
                seasonality_estimate = self.seasonality_estimation + self.trend_estimation
        else:
            seasonality_estimate = None
        fig = plot_change(
            observation=observation,
            trend_estimate=trend_estimate,
            trend_change=trend_change,
            year_seasonality_estimate=yearly_seasonality_estimate,
            adaptive_lasso_estimate=adaptive_lasso_estimate,
            seasonality_change=seasonality_change,
            seasonality_estimate=seasonality_estimate,
            yaxis=self.value_col
        )
        if fig is not None and len(fig.data) > 0:
            if plot:
                iplot(fig)
            else:
                return fig
        else:
            warnings.warn("Figure is empty, at least one component has to be true.")
            return None


def get_changepoints_dict(df, time_col, value_col, changepoints_dict):
    """The functions takes the ``changepoints_dict`` dictionary and returns the processed
    ``changepoints_dict``.

    If "method" == "auto", the change point detection algorithm is run and the function returns
    the ``changepoints_dict`` with "method"="custom" and automatically detected change points.
    If "method" == "custom" or "uniform", the original ``changepoints_dict`` is returned.

    Parameters
    ----------
    df : `pandas.DataFrame`
        The dataframe used to do change point detection.
    time_col : `str`
        The column name of time in ``df``.
    value_col : `str`
        The column name of values in ``df``.
    changepoints_dict : `dict`
        The ``changepoints_dict`` parameter that is fed into
        `~greykite.algo.forecast.silverkite.forecast_silverkite.SilverkiteForecast.forecast` or
        `~greykite.algo.forecast.silverkite.forecast_simple_silverkite.forecast_simple_silverkite`
        It must have keys:

            `"method"` : `str`, equals "custom", "uniform" or "auto"

        Depending on `"method"`, it must have keys:

            `"n_changepoints"` : `int`, when "method" == "uniform".
            `"dates"` : `Iterable[Union[int, float, str, datetime]]`, when "method" == "custom"

        When `"method"` == "auto", it can have optional keys that matches the parameters in
        `~greykite.algo.changepoint.adalasso.changepoint_detector.ChangepointDetector.find_trend_changepoints`,
        except ``df``, ``time_col`` and ``value_col``, and extra keys "dates", "combine_changepoint_min_distance"
        and "keep_detected", which correspond to the three parameters "custom_changepoint_dates", "min_distance"
        and "keep_detected" in `~greykite.algo.changepoint.adalasso.changepoints_utils.combine_detected_and_custom_trend_changepoints`.
        In all three "method" cases, it can have optional keys:

            "continuous_time_col": ``str`` or ``None``
                Column to apply `growth_func` to, to generate changepoint features
                Typically, this should match the growth term in the model
            "growth_func": ``callable`` or ``None``
                Growth function (scalar -> scalar). Changepoint features are created
                by applying `growth_func` to "continuous_time_col" with offsets.
                If None, uses identity function to use `continuous_time_col` directly
                as growth term.

    Returns
    -------
    changepoints_dict : `dict`
        If "method" == "custom" or "method" == "uniform", the return is the original dictionary.
        If "method" == "auto", the return is the dictionary with "method" = "custom" and
        "dates" = detected_changepoints. Change point detection related keys and values are used
        for change point detection, other keys and values will be included in the return dictionary.

    changepoint_detector : `ChangepointDetector` or `None`
        The ChangepointDetector class used for automatically trend changepoint detection if "method" == "auto",
        otherwise None.
    """
    if (changepoints_dict is not None
            and "method" in changepoints_dict.keys()
            and changepoints_dict["method"] == "auto"):
        changepoint_detection_args = {
            "df": df,
            "time_col": time_col,
            "value_col": value_col
        }
        changepoint_detection_keys = [
            "yearly_seasonality_order",
            "yearly_seasonality_change_freq",
            "resample_freq",
            "trend_estimator",
            "adaptive_lasso_initial_estimator",
            "regularization_strength",
            "actual_changepoint_min_distance",
            "potential_changepoint_distance",
            "potential_changepoint_n",
            "no_changepoint_distance_from_begin",
            "no_changepoint_proportion_from_end",
            "no_changepoint_distance_from_end",
            "no_changepoint_proportion_from_end"
        ]
        changepoints_dict_keys = [
            "continuous_time_col",
            "growth_func"
        ]
        for key in changepoint_detection_keys:
            if key in changepoints_dict.keys():
                changepoint_detection_args[key] = changepoints_dict[key]
        model = ChangepointDetector()
        result = model.find_trend_changepoints(**changepoint_detection_args)
        new_changepoints_dict = result["changepoints_dict"]
        custom_changepoint_keys = []
        if "dates" in changepoints_dict:
            # Here we reuse the key "dates" which is used in "method"=="custom" as additional custom dates
            # when "method"=="auto" to avoid having too many keys.
            # Gets the custom changepoints.
            custom_changepoints = changepoints_dict["dates"]
            df_dates = pd.to_datetime(df[time_col])
            min_date = min(df_dates)
            max_date = max(df_dates)
            custom_changepoints = pd.to_datetime(custom_changepoints)
            custom_changepoints = [cp for cp in custom_changepoints if min_date < cp < max_date]
            min_distance = changepoints_dict.get("combine_changepoint_min_distance", None)
            # If ``combine_changepoint_min_distance`` is not set, try to set is with ``actual_changepoint_min_distance``.
            if min_distance is None:
                min_distance = changepoints_dict.get("actual_changepoint_min_distance", None)
            # If ``keep_detected`` is not set, the default is False to keep custom changepoints.
            keep_detected = changepoints_dict.get("keep_detected", False)
            # Checks if custom changepoints are provided.
            # If provided, combines the detected changepoints and custom changepoints.
            custom_changepoint_keys = ["dates"]
            if custom_changepoints is not None:
                # Keys used in adding custom changepoints
                custom_changepoint_keys = [
                    "dates",
                    "combine_changepoint_min_distance",
                    "keep_detected"
                ]
                combined_changepoints = combine_detected_and_custom_trend_changepoints(
                    detected_changepoint_dates=new_changepoints_dict["dates"],
                    custom_changepoint_dates=custom_changepoints,
                    min_distance=min_distance,
                    keep_detected=keep_detected
                )
                # Logs info if the detected changepoints are altered.
                if not set(combined_changepoints) == set(new_changepoints_dict["dates"]):
                    log_message(f"Custom trend changepoints have been added to the detected trend changepoints."
                                f" The final trend changepoints are {combined_changepoints}",
                                LoggingLevelEnum.INFO)
                new_changepoints_dict["dates"] = combined_changepoints
        for key in changepoints_dict_keys:
            if key in changepoints_dict.keys():
                new_changepoints_dict[key] = changepoints_dict[key]
        unused_keys = [key for key in changepoints_dict.keys()
                       if key not in ["method"] + changepoints_dict_keys + changepoint_detection_keys + custom_changepoint_keys]
        if unused_keys:
            warnings.warn(f"The following keys in ``changepoints_dict`` are not recognized\n"
                          f"{unused_keys}")
        return new_changepoints_dict, model
    else:
        return changepoints_dict, None


def get_seasonality_changepoints(
        df,
        time_col,
        value_col,
        trend_changepoints_dict=None,
        trend_changepoint_dates=None,
        seasonality_changepoints_dict=None):
    """Automatically detects seasonality change points.

    The function first converts changepoints_dict if "method" == "auto", then extracts trend
    changepoint dates from the dictionary and feeds them into ``find_seasonality_changepoints``.
    With the detected seasonality change points, the detection result dictionary is returned.

    Parameters
    ----------
    df : `pandas.DataFrame`
        The data df.
    time_col : `str`
        The column name for time in ``df``.
    value_col : `str`
        The column name for value in ``df``.
    trend_changepoints_dict : `dict` or `None`, default `None`
        The ``changepoints_dict`` parameter in
        `~greykite.algo.changepoint.adalasso.changepoint_detector.get_changepoints_dict`
    trend_changepoint_dates : `list` or `None`, default `None`
        List of trend change point dates. The dates need to be parsable ty `pandas.to_datetime`.
        If given, trend change point detection will not be run and ``trend_changepoints_dict``
        will have no effect.
    seasonality_changepoints_dict : `dict` or `None`, default `None`
        The keys are the parameter names of
        `~greykite.algo.changepoint.adalasso.changepoint_detector.find_seasonality_changepoints`
        The values are the corresponding desired values.
        Note ``df``, ``time_col``, ``value_col`` and ``trend_changepoints`` are auto populated,
        and do not need to be provided.

    Returns
    -------
    result : `dict`
        The detected seasonality change points result dictionary as returned by
        `~greykite.algo.changepoint.adalasso.changepoint_detector.ChangepointDetector.find_seasonality_changepoints`.
    """
    if trend_changepoint_dates is None:
        # Runs trend change point detection algorithm if "method" == "auto", and returns
        # changepoints_dict with "method" == "custom";
        # Returns the original changepoints_dict if "method" == "uniform" or "custom";
        # Returns None if the original changepoints_dict is None.
        trend_changepoints_dict, _ = get_changepoints_dict(
            df=df,
            time_col=time_col,
            value_col=value_col,
            changepoints_dict=trend_changepoints_dict
        )
        # Extracts change point dates from the changepoints_dict.
        # Returns None if the original changepoints_dict is None.
        trend_changepoint_dates = get_changepoint_dates_from_changepoints_dict(
            changepoints_dict=trend_changepoints_dict,
            df=df,
            time_col=time_col
        )
    # Prepares arguments for ``find_seasonality_changepoints`` function.
    seasonality_changepoint_detection_args = {
        "df": df,
        "time_col": time_col,
        "value_col": value_col
    }
    seasonality_changepoint_detection_keys = [
        "seasonality_components_df",
        "resample_freq",
        "regularization_strength",
        "actual_changepoint_min_distance",
        "potential_changepoint_distance",
        "potential_changepoint_n",
        "no_changepoint_distance_from_end",
        "no_changepoint_proportion_from_end"
    ]
    if seasonality_changepoints_dict is not None:
        for key in seasonality_changepoint_detection_keys:
            if key in seasonality_changepoints_dict.keys():
                seasonality_changepoint_detection_args[key] = seasonality_changepoints_dict[key]
    seasonality_changepoint_detection_args["trend_changepoints"] = trend_changepoint_dates
    # Runs ``find_seasonality_changepoints``.
    cd = ChangepointDetector()
    result = cd.find_seasonality_changepoints(**seasonality_changepoint_detection_args)
    # Builds seasonality features with change points.
    return result



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen

import numpy as np
import pandas as pd
from sklearn.pipeline import FeatureUnion
from sklearn.pipeline import _fit_transform_one
from sklearn.pipeline import _transform_one
from sklearn.utils._joblib import Parallel
from sklearn.utils._joblib import delayed


class PandasFeatureUnion(FeatureUnion):
    """Concatenates results of multiple transformer objects.
    Transformers are expected to have pd.DataFrame as input and output

    Modified from sklearn.pipeline.FeatureUnion

    This estimator applies a list of transformer objects in parallel to the
    input data, then concatenates the results. This is useful to combine
    several feature extraction mechanisms into a single transformer.

    Parameters of the transformers may be set using its name and the parameter
    name separated by a '__'. A transformer may be replaced entirely by
    setting the parameter with its name to another transformer,
    or removed by setting to 'drop' or ``None``.

    Read more in the :ref:`User Guide <feature_union>`.

    Parameters
    ----------
    transformer_list : list of (string, transformer) tuples
        List of transformer objects to be applied to the data. The first
        half of each tuple is the name of the transformer.
        .. versionchanged:: 0.22
           Deprecated `None` as a transformer in favor of 'drop'.
    n_jobs : int, default=None
        Number of jobs to run in parallel.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors. See :term:`Glossary <n_jobs>`
        for more details.
        .. versionchanged:: v0.20
           `n_jobs` default changed from 1 to None
    transformer_weights : dict, default=None
        Multiplicative weights for features per transformer.
        Keys are transformer names, values the weights.
    verbose : bool, default=False
        If True, the time elapsed while fitting each transformer will be
        printed as it is completed.
    """
    def __init__(self, transformer_list, n_jobs=None,
                 transformer_weights=None, verbose=False):
        super().__init__(
            transformer_list=transformer_list,
            n_jobs=n_jobs,
            transformer_weights=transformer_weights,
            verbose=verbose)

    def fit_transform(self, X, y=None, **fit_params):
        """Fits all transformers, transforms the data and concatenates results.

        Modified from `sklearn.pipeline.FeatureUnion`.

        Parameters
        ----------
        X : `pandas.DataFrame`
            Input data to be transformed.

        y : array-like, shape (n_samples, ...), optional
            Targets for supervised learning.

        Returns
        -------
        X_t : `pandas.Dataframe`, shape (n_samples, sum_n_components)
            column-wise concatenation of results of transformers. sum_n_components is the
            sum of n_components (output dimension) over transformers.
        """
        results = self._parallel_func(X, y, fit_params, _fit_transform_one)

        if not results:
            # All transformers are None
            return pd.DataFrame(np.zeros((X.shape[0], 0)))
        Xs, transformers = zip(*results)
        self._update_transformer_list(transformers)
        Xs = pd.concat(Xs, axis=1)
        return Xs

    def transform(self, X):
        """Transforms X separately by each transformer, concatenates results.

        Modified from `sklearn.pipeline.FeatureUnion`

        Parameters
        ----------
        X : `pandas.DataFrame`
            Input data to be transformed.

        Returns
        -------
        X_t : `pandas.DataFrame`, shape (n_samples, sum_n_components)
            column-wise concatenation of results of transformers. sum_n_components is the
            sum of n_components (output dimension) over transformers.
        """
        Xs = Parallel(n_jobs=self.n_jobs)(
            delayed(_transform_one)(trans, X, None, weight)
            for name, trans, weight in self._iter())
        if not Xs:
            # All transformers are None
            return pd.DataFrame(np.zeros((X.shape[0], 0)))
        Xs = pd.concat(Xs, axis=1)
        return Xs



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Reza Hosseini
"""Helper functions for
`~greykite.algo.forecast.silverkite.forecast_silverkite.py.`
"""

import math
import warnings

from greykite.common.enums import SimpleTimeFrequencyEnum


def get_similar_lag(freq_in_days):
    """For a given frequency, it returns a lag which is likely to be most correlated
    to the observation at current time.

    For daily data, this will return 7 and for hourly data it will return 24*7.
    In general for sub-weekly frequencies, it returns the lag which corresponds to
    the same time in the last week.
    For data which is weekly or with frequencies larger than a week, it returns None.

    Parameters
    ----------
    freq_in_days : `float`
        The time frequency of the timeseries given in day units.

    Returns
    -------
    similar_lag : `int` or None
        The returned lag or None.
    """
    similar_lag = None
    # Get the number of observations per week
    obs_num_per_week = 7 / freq_in_days

    if obs_num_per_week > 1:
        similar_lag = math.ceil(obs_num_per_week)

    return similar_lag


def get_default_changepoints_dict(
        changepoints_method,
        num_days,
        forecast_horizon_in_days):
    """Get a changepoint dictionary based on the number of days in the observed
    timeseries and forecast horizon length in days to be provided as input to
    `~greykite.algo.forecast.silverkite.forecast_silverkite.SilverkiteForecast.forecast`.
    For the "uniform" method, we place the change points at a distance of
    ``max(28, forecast_horizon)``.
    For the "auto" method, we have used some defaults which seem to work for general
    applications::

        changepoints_dict = {
            "method": "auto",
            "yearly_seasonality_order": 10,
            "resample_freq": "7D",
            "regularization_strength": 0.8,
            "actual_changepoint_min_distance": "14D",
            "potential_changepoint_distance": "7D",
            "no_changepoint_distance_from_end": "14D"}

    If the length of data is smaller than ``2*max(28, forecast_horizon)``,
    the function will return None for all methods.

    Parameters
    ----------
    changepoints_method : `str`
        The method to locate changepoints.
        Valid options:

            - "uniform". Places changepoints evenly spaced changepoints to allow
            growth to change. The distance between the uniform change points is
            set to be ``max(28, forecast_horizon)``
            - "auto". Automatically detects change points.
            For configuration, see
            `~greykite.algo.changepoint.adalasso.changepoint_detector.ChangepointDetector.find_trend_changepoints`

        For more details for both methods, also check the documentation for
        `~greykite.algo.forecast.silverkite.forecast_silverkite.SilverkiteForecast.forecast`.
    num_days : `int`
        Number of days appearing in the observed timeseries.
    forecast_horizon_in_days : `float`
        The length of the forecast horizon in days.

    Returns
    -------
    changepoints_dict : `dict` or None
        A dictionary with change points information to be used as input to
        `~greykite.algo.forecast.silverkite.forecast_silverkite.SilverkiteForecast.forecast`.
        See that function's documentation for more details.
    """

    changepoints_dict = None
    # A reasonable distance defined based on ``forecast_horizon``
    # Here the minimum is set at 28 days
    uniform_distance = max(28, forecast_horizon_in_days)
    # Number of change points for "uniform"
    # Also if this number is zero both methods will return `None`
    changepoint_num = num_days // uniform_distance - 1

    if changepoint_num > 0:
        if changepoints_method == "uniform":
            changepoints_dict = {
                "method": "uniform",
                "n_changepoints": changepoint_num,
                "continuous_time_col": "ct1",
                "growth_func": lambda x: x}

        elif changepoints_method == "auto":
            changepoints_dict = {
                "method": "auto",
                "yearly_seasonality_order": 10,
                "resample_freq": "7D",
                "regularization_strength": 0.8,
                "actual_changepoint_min_distance": "14D",
                "potential_changepoint_distance": "7D",
                "no_changepoint_distance_from_end": "14D"}

    return changepoints_dict


def get_silverkite_uncertainty_dict(
        uncertainty,
        simple_freq=SimpleTimeFrequencyEnum.DAY.name,
        coverage=None):
    """Returns an uncertainty_dict for
    `~greykite.algo.forecast.silverkite.forecast_silverkite.SilverkiteForecast.forecast`
    input parameter: uncertainty_dict.

    The logic is as follows:

        - If ``uncertainty`` is passed as dict:
            - If ``quantiles`` are not passed through ``uncertainty`` we fill them
              using `coverage`.
            - If ``coverage`` also missing or quantiles calculated
              in two ways (via ``uncertainty["params"]["quantiles"]`` and ``coverage``)
              do not match, we throw Exceptions

        - If ``uncertainty=="auto"``:
            - We provide defaults based on time frequency of data.
            - Specify ``uncertainty["params"]["quantiles"]`` based on
              ``coverage`` if provided, otherwise the default coverage is 0.95.

    Parameters
    ----------
    uncertainty : `str` or `dict` or None
        It specifies what method should be used for uncertainty.
        If a dict is passed then it is directly returned to be passed to
        `~greykite.algo.forecast.silverkite.forecast_silverkite.SilverkiteForecast.forecast` as `uncertainty_dict`.

        If "auto", it builds a generic dict depending on frequency.
            - For frequencies less than or equal to one day it sets
              `conditional_cols` to be ["dow_hr"].
            - Otherwise it sets the conditional_cols to be `None`

        If None and `coverage` is None, the upper/lower predictions are not returned
    simple_freq : `str`, optional
        SimpleTimeFrequencyEnum member that best matches the input data frequency
        according to `get_simple_time_frequency_from_period`
    coverage : `float` or None, optional
        Intended coverage of the prediction bands (0.0 to 1.0)
        If None and `uncertainty` is None, the upper/lower predictions are not returned

    Returns
    -------
    uncertainty : `dict` or None
        An uncertainty dict to be used as input to
        `~greykite.algo.forecast.silverkite.forecast_silverkite.SilverkiteForecast.forecast`.
        See that function's docstring for more details.
    """
    frequency = SimpleTimeFrequencyEnum[simple_freq].value

    # boolean to determine if freq is longer than one day
    freq_is_longer_than_day = (
            frequency.seconds_per_observation
            > SimpleTimeFrequencyEnum.DAY.value.seconds_per_observation)

    uncertainty_dict = None

    # if both `uncertainty` and `coverage` are None, we return None
    if uncertainty is None and coverage is None:
        return None

    # checking if coverage input is sensible
    if coverage is not None and (coverage < 0 or coverage > 1):
        raise ValueError("coverage must be between 0 and 1")

    # if only coverage is provided, consider uncertainty to be "auto"
    if coverage is not None and uncertainty is None:
        uncertainty = "auto"

    # The case where `uncertainty` is input as a dict
    # We check if quantiles are passed through `uncertainty`
    # If not, we use `coverage` to fill them in
    # If quantiles are passed in `uncertainty` and inferrable from `coverage`:
    # and they are inconsistent, we throw an Exception
    if isinstance(uncertainty, dict):
        uncertainty_dict = uncertainty
        # boolean to check if quantiles are passed through uncertainty
        try:
            quantiles_specified = (uncertainty["params"]["quantiles"] is not None)
        except KeyError:
            quantiles_specified = False
        if "params" not in uncertainty_dict:
            uncertainty_dict["params"] = {}

        if quantiles_specified:
            quantiles = uncertainty["params"]["quantiles"]
            # If quantiles are specified, we do some sanity checks on their values:
            # We give warnings if more than two quantiles were passed
            # or if they are not symmetric i.e. first quantiles distance to zero
            # is not the same as last quantile distance to 1
            # We throw exceptions if quantiles are not increasing
            # or if `coverage` is also passed and inconsistent with `quantiles`
            if len(quantiles) > 2:
                warnings.warn(
                    "More than two quantiles are passed in `uncertainty`."
                    " Confidence intervals will be based on"
                    " the first (lower limit) and last (upper limit) quantile",
                    Warning)
            coverage_via_uncertainty = quantiles[-1] - quantiles[0]
            if coverage_via_uncertainty <= 0:
                raise ValueError(
                    "`quantiles` is expected to be an increasing sequence"
                    " of at least two elements."
                    f"These quantiles were passed: quantiles = {quantiles}")
            if round(quantiles[-1], 3) != round(1 - quantiles[0], 3):
                warnings.warn(
                    "1 - (quantiles upper limit) is not equal to (quantiles lower limit)"
                    " (lack of symmetry)."
                    f" Asymmetric quantiles: {quantiles} were used.",
                    Warning)
            if coverage is not None:
                # The case where quantiles are both provided through `uncertainty`
                # and inferrable using `coverage`
                # We check for conflict in coverage specification
                if round(coverage_via_uncertainty, 3) != round(coverage, 3):
                    raise ValueError(
                        "Coverage is specified/inferred both via `coverage` and via `uncertainty` input"
                        " and values do not match."
                        f" Coverage specified via `coverage`: {round(coverage, 3)}."
                        f" Coverage inferred via `uncertainty`: {round(coverage_via_uncertainty, 2)}.")
        if not quantiles_specified:
            if coverage is None:
                raise ValueError(
                    "`quantiles` are not specified in `uncertainty`"
                    " and `coverage` is not provided to infer them")
            else:
                # The case where quantiles is not provided through `uncertainty`
                # but coverage is passed
                q1 = (1 - coverage)/2
                q2 = 1 - q1
                uncertainty_dict["params"]["quantiles"] = [q1, q2]

    # The case where `uncertainty` is passed as "auto"
    # The auto case conditions data on `dow_hr` which represents day of week and hour
    # for data with frequency less than or equal to a day (e.g. hourly, daily)
    # note that for daily case this works too as dow_hr will only depend on dow
    if uncertainty == "auto":
        if not freq_is_longer_than_day:
            uncertainty_dict = {
                "uncertainty_method": "simple_conditional_residuals",
                "params": {
                    "conditional_cols": ["dow_hr"],
                    "quantiles": [0.025, 0.975],
                    "quantile_estimation_method": "normal_fit",
                    "sample_size_thresh": 5,
                    "small_sample_size_method": "std_quantiles",
                    "small_sample_size_quantile": 0.98}}
        else:
            uncertainty_dict = {
                "uncertainty_method": "simple_conditional_residuals",
                "params": {
                    "conditional_cols": None,
                    "quantiles": [0.025, 0.975],
                    "quantile_estimation_method": "normal_fit",
                    "sample_size_thresh": 5,
                    "small_sample_size_method": "std_quantiles",
                    "small_sample_size_quantile": 0.98}}
        # if coverage is provided the quantiles are overridden in auto
        # we do not give warnings as it is the auto case and
        # user expects using the coverage provided
        if coverage is not None:
            q1 = (1 - coverage)/2
            q2 = 1 - q1
            uncertainty_dict["params"]["quantiles"] = [q1, q2]

    return uncertainty_dict



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Reza Hosseini
"""Functions to construct lagged time series features."""
import numpy as np
import pandas as pd

from greykite.common import constants as cst


def build_lag_df(
        value_col,
        df=None,
        max_order=None,
        orders=None):
    """A function which builds a dataframe including time series lags
    (in the form of data frame columns)
    for a given value column (value_col) from an input dataframe (df)

    :param value_col: str
        the column name for the column which includes the values
    :param df: Optional[pd.DataFrame]
        data frame which include the value column of interest
        df could be passed as None if col_names are desired only
        in the output dictionary
    :param max_order: Optional[int]
        if lag orders are not specified then max_order is used.
        max_order is a natural number specifying
        the orders needed. e.g. if "max_order = 5"
        then we add these lag orders 1, 2, 3, 4, 5
    :param orders: List[int]
        a list of the lag orders needed.
        e.g. if orders = [1, 2, 7] for a given time series
        denoted in mathematical notation by Y(t), we calculate
        Y(t-1), Y(t-2), Y(t-7) and store them in the returned data frame
    :return: dict
        dictionary with these items:

            - "col_names": List[str]
                the generated column names
            - "lag_df": Optional[pd.DataFrame]
                a data frame consisting of the lagged values of the given orders.
                For example, if the value_col = "y" and the orders = [1, 2, 7],
                then the returned data frame
                include the following columns: "y_lag_1", "y_lag_2" and "y_lag_7".
                These correspond to the 1st lag, the 2nd lag and the 7th lag
                respectively.
    """
    # initializes the returned items
    lag_df = None
    col_names = []

    if max_order is None and orders is None:
        raise ValueError(
            "at least one of 'max_order' or 'orders' must be provided")

    if orders is None:
        orders = range(1, max_order + 1)

    if df is not None:
        lag_df = pd.DataFrame()

    for i in orders:
        col_name = f"{value_col}{cst.LAG_INFIX}{i}"
        col_names.append(col_name)
        if df is not None:
            lag_df[col_name] = df[value_col].shift(i)
    return {
        "lag_df": lag_df,
        "col_names": col_names}


def min_max_lag_order(
        lag_dict=None,
        agg_lag_dict=None):
    """Calculating min and max lag order needed given the prescribed lags
    in the model.

    :param lag_dict: dict
        dictionary with these fields:

            - "orders": Optional[List(int)]
             a list of positive integers or None
            - "max_order": Optional[int]

        see function "build_lag_df" arguments with same names
        to understand use case
    :param agg_lag_dict: dict
        dictionary with these items

            - "orders_list": List[List[int]], default: []
                a list of lists of integers
            - "interval_list": List[tuple[int]], default: []
                a list of int tuples of each with length 2

        see function "build_agg_lag_df" arguments with same names
        to understand use-case

    :return: dict[str, float]
        dictionary with two items with keys:

            - "max_order": the maximum lag used
            - "min_order": the mimimum lag used
    """
    max_order = 0
    min_order = np.inf

    if lag_dict is not None:
        if lag_dict.get("orders") is not None:
            max_order = np.nanmax(lag_dict["orders"])
            min_order = np.nanmin(lag_dict["orders"])
        elif lag_dict.get("max_order") is not None:
            max_order = np.nanmax([lag_dict["max_order"], max_order])
            # if max_order is not None in `lag_dict`
            # then lags 1, 2, ..., `max_order` are used
            # Therfore `min_order` is 1
            min_order = 1

    if agg_lag_dict is not None:
        orders_list = []
        if agg_lag_dict.get("orders_list") is not None:
            orders_list = orders_list + agg_lag_dict.get("orders_list")
        if agg_lag_dict.get("interval_list") is not None:
            orders_list = orders_list + agg_lag_dict.get("interval_list")
        flatten_orders = [order for sublist in orders_list for order in sublist]
        max_order = np.nanmax(flatten_orders + [max_order])
        min_order = np.nanmin(flatten_orders + [min_order])

    return {
        "max_order": max_order,
        "min_order": min_order}


def build_agg_lag_df(
        value_col,
        df=None,
        orders_list=[],
        interval_list=[],
        agg_func=np.mean,
        agg_name=cst.AGG_LAG_INFIX,
        max_order=None):
    """A function which returns a dataframe including aggregated
    (e.g. averaged) time series lags in the form of dataframe columns.
    By "aggregated lags", we mean an aggregate of several lags using an
    aggregation function given in "agg_func".
    The advantage of "aggregated lags" over regular lags is we can aggregate
    (e.g. average) many lags in the past instead of using a large number of lags.
    This is useful in many applications and avoids over-fitting.

    For a time series mathematically denoted by Y(t),
    one could consider the average lag processes as follows:
        the average of last 3 values:
            "avg(t) = (Y(t-1) + Y(t-2) + Y(t-3)) / 3"
        the average of 7th, 14th and 21st lags:
            "avg(t) = (Y(t-7) + Y(t-14) + Y(t-21)) / 3"

    See following references:
        Reza Hosseini et al. (2014)
        Non-linear time-varying stochastic models for agroclimate risk assessment,
        Environmental and Ecological Statistics
        https://link.springer.com/article/10.1007/s10651-014-0295-2

        Alireza Hosseini et al. (2017)
        Capturing the time-dependence in the precipitation process for weather risk assessment,
        Stochastic Environmental Research and Risk Assessment
        https://link.springer.com/article/10.1007/s00477-016-1285-8

    :param value_col: str
        the column name for the values of interest
    :param df: Optional[pd.DataFrame]
        the data frame which includes the time series of interest
    :param orders_list: List[int]
        a list including the order range for the average lags. For example if
        agg_func = np.mean and orders_list = [[1, 2, 3], [7, 14, 21]]
        then we construct two averaged lags:
            avg(t) = (Y(t-1) + Y(t-2) + Y(t-3)) / 3 and
            avg(t) = (Y(t-7) + Y(t-14) + Y(t-21)) / 3
    :param interval_list: List[tuple[int]]
        a list of (lag) intervals
        where interval is a tuple of length 2 with
            - first element denoting the lower bound and
            - second is the upper
        For example if interval_list = [(1, 3), (8, 11)]
        then we construct two "average lagged" variables:
            avg(t) = (Y(t-1) + Y(t-2) + Y(t-3)) / 3 and
            avg(t) = (Y(t-8) + Y(t-9) + Y(t-10) + Y(t-11)) / 4
    :param agg_func: callable, default: np.mean
        the function used to aggregate the lag orders for each of
        orders specified in either of order_list or interval_list.
        Typically this function is an averaging function such as
        np.mean or np.median but more sophisticated functions are allowed.
    :param agg_name: str, default: "avglag"
        the aggregate function name used in constructing the column names for
        the output data frame.
        For example if
            - value_col = "y"
            - orders = [7 , 14, 21]
            - agg_name = "avglag"
        then the column name appearing in the output data frame
        will be "y_avglag_7_14_21".
    :param max_order: Optional[int]
        maximum order of lags needed in calculations of lag aggregates
        this is usually calculated/inferred from these arguments:
            orders_list, interval_list
        unless the max_order is already pre-calculated before calling
        this function. Hence this argument is optional and only included for
        computational efficiency gains.
    :return: dict
        dictionary with following items:
            - "col_names": List[str]
                the generated column names
            - "agg_lag_df": Optional[pd.DataFrame]
                a data frame with the average lag columns.
                The column names are constructed in a way
                that reflects what lags are averaged. For example if
                    - value_col = "y"
                    - agg_name = "avglag"
                    - orders_list = [[1, 2, 3], [7, 14, 21]]
                Then the column names are
                "y_avglag_1_2_3", "y_avglag_7_14_21"
                and if
                    - interval_list = [(1, 3), (8, 11)]
                Then the column names are
                "y_avglag_1_to_3", "y_avglag_8_to_11"
    """

    if orders_list is None and interval_list is None:
        raise ValueError(
            "at least one of 'orders_list' or 'interval_list' must be provided")
    # find out which lags we need by finding the maximum lag used
    # note that max_order should be usually passed as None (default)
    # unless it is pre-calculated before calling this function
    if max_order is None:
        max_order = min_max_lag_order(
            lag_dict=None,
            agg_lag_dict={
                "orders_list": orders_list,
                "interval_list": interval_list})["max_order"]

    # intialize the returned items
    agg_lag_df = None
    col_names = []

    if df is not None:
        lag_info = build_lag_df(
            df=df.copy(),
            value_col=value_col,
            max_order=max_order,
            orders=None)
        lag_df = lag_info["lag_df"]
        agg_lag_df = pd.DataFrame()

    for orders in orders_list:
        if len(orders) > len(set(orders)):
            raise Exception(
                "a list of orders in orders_list contains a duplicate element")
        col_suffix = "_".join([str(x) for x in orders])
        orders_col_index = [x-1 for x in orders]
        col_name = f"{value_col}_{agg_name}_{col_suffix}"
        col_names.append(col_name)
        if df is not None:
            agg_lag_df[col_name] = (
                lag_df.iloc[:, orders_col_index].apply(agg_func, axis=1))

    for interval in interval_list:
        if len(interval) != 2:
            raise Exception("interval must be a tuple of length 2")
        lower = interval[0]
        upper = interval[1]
        if lower > upper:
            raise Exception(
                "we must have interval[0] <= interval[1], "
                "for each interval in interval_list")
        orders = range(lower, upper + 1)
        col_suffix = f"{lower}_to_{upper}"
        orders_col_index = [x-1 for x in orders]
        col_name = f"{value_col}_{agg_name}_{col_suffix}"
        col_names.append(col_name)
        if df is not None:
            agg_lag_df[col_name] = (
                lag_df.iloc[:, orders_col_index].apply(agg_func, axis=1))

    return {
        "agg_lag_df": agg_lag_df,
        "col_names": col_names}


def build_autoreg_df(
        value_col,
        lag_dict=None,
        agg_lag_dict=None,
        series_na_fill_func=lambda s: s.bfill().ffill()):
    """This function generates a function ("build_lags_func" in the returned dict)
        which when called builds a lag data frame and an aggregated lag data frame using
        "build_lag_df" and "build_agg_lag_df" functions.
        Note: In case of training, validation and testing (e.g. cross-validation)
        for forecasting, this function needs to be applied after the data split is done.
        This is especially important if "series_na_fill_func" is using future values
        in interpolation - that is the case for the default which is
        lambda s: s.bfill().ffill()

   :param value_col: str
        the column name for the values of interest
   :param lag_dict: Optional[dict]
        A dictionary which encapsulates the needed params to be passed to the
        function "build_lag_df"
        Expected items are:

            - "max_order": Optional[int]
                the max_order for creating lags
            - "orders": Optional[List[int]]
                the orders for which lag is needed

    :param agg_lag_dict: Optional[dict]
        A dictionary encapsulating the needed params to be passed to the function
        "build_agg_lag_df"
        Expected items are:

            - "orders_list": List[List[int]]
                A list of list of integers.
                Each int list is to be used as order of lags to be aggregated
                See build_lag_df arguments for more details
            - "interval_list": List[tuple]
                A list of tuples each of length 2.
                Each tuple is used to construct an aggregated lag using all orders within that range
                See build_agg_lag_df arguments for more details
            - "agg_func": func (pd.Dataframe -> pd.Dataframe)
                The function used for aggregation in "build_agg_lag_df"
                If this key is not passed, the default of "build_agg_lag_df" will be used

    :param series_na_fill_func: (pd.Series -> pd.Series)
        default: lambda s: s.bfill.ffill()
        This function is used to fill in the missing data
        The default works by first back-filling and then forward-filling
        This function should not be applied to data before CV split is done.
    :return: dict
        a dictionary with following items

            - "build_lags_func": func
                pd.Daframe -> dict(lag_df=pd.DataFrame, agg_lag_df=pd.DataFrame)
                A function which takes a df (need to have value_col) as input
                calculates the lag_df and agg_lag_df and returns them
            - "lag_col_names": Optional[List[str]]
                The list of generated column names for the returned lag_df
                when "build_lags_func" is applied
            - "agg_lag_col_names": Optional[List[str]]
                The list of generated column names for returned agg_lag_df when
                "build_lags_func" is applied
            - "max_order": int
                the maximum lag order needed in the calculation of "build_lags_func"
            - "min_order": int
                the minimum lag order needed in the calculation of "build_lags_func"

    """
    # building arguments for passing to build_lag_df
    # when lag_dict is not None
    build_lag_df_args = None
    if lag_dict is not None:
        build_lag_df_args = {"value_col": value_col}
        build_lag_df_args.update(lag_dict)

    # building arguments for passing to build_agg_lag_df
    # when agg_lag_dict is not None
    build_agg_lag_df_args = None
    if agg_lag_dict is not None:
        build_agg_lag_df_args = {"value_col": value_col}
        build_agg_lag_df_args.update(agg_lag_dict)

    # we get the col_names for lag_df
    lag_col_names = None
    if lag_dict is not None:
        lag_info = build_lag_df(
            df=None,
            **build_lag_df_args)
        lag_col_names = lag_info["col_names"]

    # we get col_names for agg_lag_df
    agg_lag_col_names = None
    if agg_lag_dict is not None:
        agg_lag_info = build_agg_lag_df(
            df=None,
            **build_agg_lag_df_args)
        agg_lag_col_names = agg_lag_info["col_names"]

    # we find out the max_order needed
    # outside the internal function: build_lags_func
    min_max_order = min_max_lag_order(
        lag_dict=build_lag_df_args,
        agg_lag_dict=build_agg_lag_df_args)
    max_order = min_max_order["max_order"]
    min_order = min_max_order["min_order"]

    def build_lags_func(df, past_df=None):
        """A function which uses:
            df (pd.Dataframe), past_df (pd.DataFrame)
            and returns lag_df and agg_lag_df for df.
            This function infers some parameters
            e.g. value_col, max_order, series_na_fill_func from its environment.
        :param df: pd.DataFrame
             The input dataframe which is expected to have value_col as a column
             The returned lag_df and agg_lag_df are generated for df
             (past_df will be used in calculation if provided)
        :param past_df: Optional[pd.DataFrame]
            When provided it will be appended to df (from left)
            in order for the lags to be calculated.
            past_df is considered to include the past values for the time series
            leading up to df values. So if df values start at time t0:
                Y(t0), ..., Y(t0 + len(df))
            past_df will include these values
                Y(t-len(past_df)), ..., Y(t0-1)
            Note that the last value in past_df is the immediate value in time
            before t0 which is the first time in df.
            Also note that we do not require a timestamp column in df and past_df
            as that is not needed in the logic.
            If past_df is None, and series_na_fill_func is also None.
            Therefore lag_df and agg_lag_df will include NULLs at the beginning
            depending on how many lags are calculated
        :return: dict
            A dictionary including lag_df and agg_lag_df for the input df
            The dictionary includes following items:
                - "lag_df": Optional[pd.DataFrame]
                - "agg_lag_df": Optional[pd.DataFrame]
        """
        df = df[[value_col]].reset_index(drop=True)

        # if past_df is None, we create one with np.nan
        # also if it is shorter than max_order we expand it with np.nan
        if past_df is None:
            past_df = pd.DataFrame({value_col: [np.nan]*max_order})
        else:
            if value_col not in list(past_df.columns):
                raise ValueError(
                    f"{value_col} must appear in past_df if past_df is not None")
            past_df = past_df[[value_col]].reset_index(drop=True)
            # if past_df length (number of rows) is smaller than max_order
            # we expand it to avoid NULLs
            if past_df.shape[0] < max_order:
                past_df_addition = pd.DataFrame(
                    {value_col: [np.nan]*(max_order - past_df.shape[0])})
                past_df = past_df_addition.append(past_df)

        # df is expanded by adding past_df as the past data for df
        # this will help in avoiding NULLs to appear in lag_df and agg_lag_df
        # as long as past_df has data in it or expanded df is interpolated
        df_expanded = past_df.append(df)
        if series_na_fill_func is not None:
            df_expanded[value_col] = series_na_fill_func(df_expanded[value_col])

        # we get the col_names for lag_df
        lag_df = None
        if lag_dict is not None:
            lag_info = build_lag_df(
                df=df_expanded,
                **build_lag_df_args)
            # since the lag calculation is done on expanded dataframe (`df_expanded`)
            # we need to pick only the relevant rows which match original
            # dataframe (via `iloc`)
            lag_df = lag_info["lag_df"].iloc[-(df.shape[0]):].reset_index(
                drop=True)

        # we get col_names for agg_lag_df
        agg_lag_df = None
        if agg_lag_dict is not None:
            agg_lag_info = build_agg_lag_df(
                df=df_expanded,
                **build_agg_lag_df_args)
            # since the lag calculation is done on expanded dataframe (`df_expanded`)
            # we need to pick only the relevant rows which match original
            # dataframe (via `iloc`)
            agg_lag_df = agg_lag_info["agg_lag_df"].iloc[
                -(df.shape[0]):].reset_index(drop=True)

        return {
            "lag_df": lag_df,
            "agg_lag_df": agg_lag_df
        }

    return {
        "build_lags_func": build_lags_func,
        "lag_col_names": lag_col_names,
        "agg_lag_col_names": agg_lag_col_names,
        "min_order": min_order,
        "max_order": max_order
    }


def build_autoreg_df_multi(
        value_lag_info_dict,
        series_na_fill_func=lambda s: s.bfill().ffill()):
    """A function which returns a function to build autoregression
    dataframe for multiple value columns.
    This function should not be applied to data before CV split is done.


    Parameters
    ----------
    value_lag_info_dict : `dict` [`str`, `dict`]
        A dictionary with keys being the target value columns: `value_col`
        For each of these value columns, a dictionary with following keys

            `lag_dict`,
            `agg_lag_dict`,
            `series_na_fill_func`

        The `value_col` and the above three variables are then passed to the
        following function:

            build_autoreg_df(
                value_col,
                lag_dict,
                agg_lag_dict,
                series_na_fill_func)

        Check the
        `greykite.common.features.timeseries_lags.build_autoreg_df`
        docstring for more details for each argument.
    series_na_fill_func : callable, (pd.Series -> pd.Series)
        default: `lambda s: s.bfill.ffill()`
        This function is used to fill in the missing data
        The default works by first back-filling and then forward-filling

    Returns
    -------
    A dictionary with following items

        "autoreg_func" : callable, (pd.DataFrame -> pd.DataFrame)
            A function which can be applied to a dataframe and return a dataframe
            which has the lagged values for all the relevant columns
        "autoreg_col_names" : List[str]
            A list of all the generated columns

    """

    multi_autoreg_info = {}
    autoreg_col_names = []
    for value_col, lag_info in value_lag_info_dict.items():
        # we assign the interpolation function to be the default specified above
        # in this function
        # if a custom interpolation function is made available for that
        # `value_col`, we replace the default
        series_na_fill_func0 = lag_info.get(
            "series_na_fill_func",
            series_na_fill_func)

        autoreg_info = build_autoreg_df(
            value_col,
            lag_dict=lag_info.get("lag_dict"),
            agg_lag_dict=lag_info.get("agg_lag_dict"),
            series_na_fill_func=series_na_fill_func0)

        # store the result (`autoreg_info`) for each `value_col` in a dictionary
        # the result for each `value_col` will be
        # a dictionary with following keys:
        # "build_lags_func"
        # "lag_col_names"
        # "agg_lag_col_names"
        # "min_order"
        # "max_order"
        multi_autoreg_info[value_col] = autoreg_info

        # extract column names for lagged variables and add to the full list of
        # all lagged variable column names: `autoreg_col_names`
        lag_col_names = autoreg_info["lag_col_names"]
        agg_lag_col_names = autoreg_info["agg_lag_col_names"]
        if lag_col_names is not None:
            autoreg_col_names += lag_col_names
        if agg_lag_col_names is not None:
            autoreg_col_names += agg_lag_col_names

    def build_lags_func_multi(df, past_df=None):
        """A function which generates a lagged dataframe
        for a given dataframe with potentially multiple value columns
        to be lagged. Note `df` and `past_df` must have the same extract
        columns. However time column is not needed as the function assumes
        `past_df` is the data which precedes `df` (without gaps).

        Parameters
        ----------
        df : `pd.DataFrame`
            The dataframe which includes the value columns for which lagged
            data is needed
        past_df : `pandas.DataFrame` or None, default None
            The past data which is immediately before `df` with same
            value columns.


        Returns :
        -------
        autoreg_df : `pandas.DataFrame`
            A DataFrame which includes the lagged values as columns
        """

        autoreg_df = None
        for value_col, lag_info in multi_autoreg_info.items():
            build_lags_func = lag_info["build_lags_func"]
            res = build_lags_func(
                df=df[[value_col]],
                past_df=past_df)
            lag_df = res["lag_df"]
            agg_lag_df = res["agg_lag_df"]
            if lag_df is not None or agg_lag_df is not None:
                # concatenating the new generated new columns to the result
                # dataframe  in column-wise fashion
                autoreg_df = pd.concat(
                    [autoreg_df, lag_df, agg_lag_df],
                    axis=1)

        return autoreg_df

    return {
        "autoreg_func": build_lags_func_multi,
        "autoreg_col_names": autoreg_col_names}



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen
"""Template for `greykite.sklearn.estimator.silverkite_estimator`.
Takes input data and forecast config,
and returns parameters to call
:func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`.
"""

import dataclasses
import functools
from typing import Dict
from typing import Optional

import pandas as pd

from greykite.algo.forecast.silverkite.forecast_silverkite import SilverkiteForecast
from greykite.algo.forecast.silverkite.silverkite_diagnostics import SilverkiteDiagnostics
from greykite.common.python_utils import dictionaries_values_to_lists
from greykite.common.python_utils import unique_in_list
from greykite.common.python_utils import update_dictionaries
from greykite.common.python_utils import update_dictionary
from greykite.framework.templates.autogen.forecast_config import ForecastConfig
from greykite.framework.templates.autogen.forecast_config import ModelComponentsParam
from greykite.framework.templates.base_template import BaseTemplate
from greykite.sklearn.estimator.base_forecast_estimator import BaseForecastEstimator
from greykite.sklearn.estimator.silverkite_estimator import SilverkiteEstimator


def get_extra_pred_cols(model_components=None):
    """Gets extra predictor columns from the model components for
    :func:`~greykite.framework.templates.silverkite_templates.silverkite_template`.

    Parameters
    ----------
    model_components : :class:`~greykite.framework.templates.autogen.forecast_config.ModelComponentsParam` or None, default None
        Configuration of model growth, seasonality, events, etc.
        See :func:`~greykite.framework.templates.silverkite_templates.silverkite_template`
        for details.

    Returns
    -------
    extra_pred_cols : `list` [`str`]
        All extra predictor columns used in any hyperparameter set
        requested by ``model_components.custom["extra_pred_cols]``.
        Regressors are included in this list.
        None if there are no extra predictor columns.
    """
    if model_components is not None and model_components.custom is not None:
        # ``extra_pred_cols`` is a list of strings to initialize
        # SilverkiteEstimator.extra_pred_cols, or a list of
        # such lists.
        extra_pred_cols = model_components.custom.get("extra_pred_cols", [])
    else:
        extra_pred_cols = []
    return unique_in_list(
        array=extra_pred_cols,
        ignored_elements=(None,))


def apply_default_model_components(
        model_components=None,
        time_properties=None):
    """Sets default values for ``model_components``.

    Parameters
    ----------
    model_components : :class:`~greykite.framework.templates.autogen.forecast_config.ModelComponentsParam` or None, default None
        Configuration of model growth, seasonality, events, etc.
        See :func:`~greykite.framework.templates.silverkite_templates.silverkite_template` for details.
    time_properties : `dict` [`str`, `any`] or None, default None
        Time properties dictionary (likely produced by
        `~greykite.common.time_properties_forecast.get_forecast_time_properties`)
        with keys:

        ``"period"`` : `int`
            Period of each observation (i.e. minimum time between observations, in seconds).
        ``"simple_freq"`` : `SimpleTimeFrequencyEnum`
            ``SimpleTimeFrequencyEnum`` member corresponding to data frequency.
        ``"num_training_points"`` : `int`
            Number of observations for training.
        ``"num_training_days"`` : `int`
            Number of days for training.
        ``"start_year"`` : `int`
            Start year of the training period.
        ``"end_year"`` : `int`
            End year of the forecast period.
        ``"origin_for_time_vars"`` : `float`
            Continuous time representation of the first date in ``df``.

    Returns
    -------
    model_components : :class:`~greykite.framework.templates.autogen.forecast_config.ModelComponentsParam`
        The provided ``model_components`` with default values set
    """
    if model_components is None:
        model_components = ModelComponentsParam()
    else:
        # makes a copy to avoid mutating input
        model_components = dataclasses.replace(model_components)

    # sets default values
    default_seasonality = {
        "fs_components_df": [pd.DataFrame({
            "name": ["tod", "tow", "tom", "toq", "toy"],
            "period": [24.0, 7.0, 1.0, 1.0, 1.0],
            "order": [3, 3, 1, 1, 5],
            "seas_names": ["daily", "weekly", "monthly", "quarterly", "yearly"]})],
    }
    model_components.seasonality = update_dictionary(
        default_seasonality,
        overwrite_dict=model_components.seasonality,
        allow_unknown_keys=False)

    # model_components.growth must be empty.
    # Pass growth terms via `extra_pred_cols` instead.
    default_growth = {}
    model_components.growth = update_dictionary(
        default_growth,
        overwrite_dict=model_components.growth,
        allow_unknown_keys=False)

    default_events = {
        "daily_event_df_dict": [None],
    }
    model_components.events = update_dictionary(
        default_events,
        overwrite_dict=model_components.events,
        allow_unknown_keys=False)

    default_changepoints = {
        "changepoints_dict": [None],
        "seasonality_changepoints_dict": [None],
        # Not allowed, to prevent leaking future information
        # into the past. Pass `changepoints_dict` with method="auto" for
        # automatic detection.
        # "changepoint_detector": [None],
    }
    model_components.changepoints = update_dictionary(
        default_changepoints,
        overwrite_dict=model_components.changepoints,
        allow_unknown_keys=False)

    default_autoregression = {
        "autoreg_dict": [None],
    }
    model_components.autoregression = update_dictionary(
        default_autoregression,
        overwrite_dict=model_components.autoregression,
        allow_unknown_keys=False)

    default_regressors = {}
    model_components.regressors = update_dictionary(
        default_regressors,
        overwrite_dict=model_components.regressors,
        allow_unknown_keys=False)

    default_uncertainty = {
        "uncertainty_dict": [None],
    }
    model_components.uncertainty = update_dictionary(
        default_uncertainty,
        overwrite_dict=model_components.uncertainty,
        allow_unknown_keys=False)

    if time_properties is not None:
        origin_for_time_vars = time_properties.get("origin_for_time_vars")
    else:
        origin_for_time_vars = None

    default_custom = {
        "silverkite": [SilverkiteForecast()],  # NB: sklearn creates a copy in grid search
        "silverkite_diagnostics": [SilverkiteDiagnostics()],
        # The same origin for every split, based on start year of full dataset.
        # To use first date of each training split, set to `None` in model_components.
        "origin_for_time_vars": [origin_for_time_vars],
        "extra_pred_cols": ["ct1"],  # linear growth
        "fit_algorithm_dict": [{
            "fit_algorithm": "linear",
            "fit_algorithm_params": None,
        }],
        "min_admissible_value": [None],
        "max_admissible_value": [None],
    }
    model_components.custom = update_dictionary(
        default_custom,
        overwrite_dict=model_components.custom,
        allow_unknown_keys=False)

    # sets to {} if None, for each item if
    # `model_components.hyperparameter_override` is a list of dictionaries
    model_components.hyperparameter_override = update_dictionaries(
        {},
        overwrite_dicts=model_components.hyperparameter_override)

    return model_components


class SilverkiteTemplate(BaseTemplate):
    """A template for :class:`~greykite.sklearn.estimator.silverkite_estimator.SilverkiteEstimator`.

    Takes input data and optional configuration parameters
    to customize the model. Returns a set of parameters to call
    :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`.

    Notes
    -----
    The attributes of a `~greykite.framework.templates.autogen.forecast_config.ForecastConfig` for
    :class:`~greykite.sklearn.estimator.silverkite_estimator.SilverkiteEstimator` are:

        computation_param: `ComputationParam` or None, default None
            How to compute the result. See
            :class:`~greykite.framework.templates.autogen.forecast_config.ComputationParam`.
        coverage: `float` or None, default None
            Intended coverage of the prediction bands (0.0 to 1.0).
            Same as coverage in ``forecast_pipeline``.
            You may tune how the uncertainty is computed
            via `model_components.uncertainty["uncertainty_dict"]`.
        evaluation_metric_param: `EvaluationMetricParam` or None, default None
            What metrics to evaluate. See
            :class:`~greykite.framework.templates.autogen.forecast_config.EvaluationMetricParam`.
        evaluation_period_param: `EvaluationPeriodParam` or None, default None
            How to split data for evaluation. See
            :class:`~greykite.framework.templates.autogen.forecast_config.EvaluationPeriodParam`.
        forecast_horizon: `int` or None, default None
            Number of periods to forecast into the future. Must be > 0
            If None, default is determined from input data frequency
            Same as forecast_horizon in `forecast_pipeline`
        metadata_param: `MetadataParam` or None, default None
            Information about the input data. See
            :class:`~greykite.framework.templates.autogen.forecast_config.MetadataParam`.
        model_components_param: `ModelComponentsParam` or None, default None
            Parameters to tune the model. See
            :class:`~greykite.framework.templates.autogen.forecast_config.ModelComponentsParam`.
            The fields are dictionaries with the following items.

            See inline comments on which values accept lists for grid search.

            seasonality: `dict` [`str`, `any`] or None, optional
                How to model the seasonality. A dictionary with keys corresponding to
                parameters in `~greykite.algo.forecast.silverkite.forecast_silverkite.SilverkiteForecast.forecast`.

                Allowed keys: ``"fs_components_df"``.
            growth: `dict` [`str`, `any`] or None, optional
                How to model the growth.

                Allowed keys: None. (Use ``model_components.custom["extra_pred_cols"]`` to specify
                growth terms.)
            events: `dict` [`str`, `any`] or None, optional
                How to model the holidays/events. A dictionary with keys corresponding to
                parameters in `~greykite.algo.forecast.silverkite.forecast_silverkite.SilverkiteForecast.forecast`.

                Allowed keys: ``"daily_event_df_dict"``.

                .. note::

                    Event names derived from ``daily_event_df_dict`` must be specified via
                    ``model_components.custom["extra_pred_cols"]`` to be included in the model.
                    This parameter has no effect on the model unless event names are passed to
                    ``extra_pred_cols``.

                    The function
                    `~greykite.algo.forecast.silverkite.forecast_simple_silverkite_helper.get_event_pred_cols`
                    can be used to extract all event names from ``daily_event_df_dict``.

            changepoints: `dict` [`str`, `any`] or None, optional
                How to model changes in trend and seasonality. A dictionary with keys corresponding to
                parameters in `~greykite.algo.forecast.silverkite.forecast_silverkite.SilverkiteForecast.forecast`.

                Allowed keys: "changepoints_dict", "seasonality_changepoints_dict", "changepoint_detector".
            autoregression: `dict` [`str`, `any`] or None, optional
                Specifies the autoregression configuration. Dictionary with the following optional key:

                ``"autoreg_dict"`` : `dict` or None or a list of such values for grid search
                    A dictionary with arguments for `~greykite.common.features.timeseries_lags.build_autoreg_df`.
                    That function's parameter ``value_col`` is inferred from the input of
                    current function ``forecast_silverkite``. Other keys are:

                        * ``"lag_dict"`` : `dict` or None
                        * ``"agg_lag_dict"`` : `dict` or None
                        * ``"series_na_fill_func"`` : callable

                    See more details for above parameters in
                    `~greykite.common.features.timeseries_lags.build_autoreg_df`.

            regressors: `dict` [`str`, `any`] or None, optional
                How to model the regressors.

                Allowed keys: None. (Use ``model_components.custom["extra_pred_cols"]`` to specify
                regressors.)
            uncertainty: `dict` [`str`, `any`] or None, optional
                How to model the uncertainty. A dictionary with keys corresponding to
                parameters in `~greykite.algo.forecast.silverkite.forecast_silverkite.SilverkiteForecast.forecast`.

                Allowed keys: ``"uncertainty_dict"``.
            custom: `dict` [`str`, `any`] or None, optional
                Custom parameters that don't fit the categories above. A dictionary with keys corresponding to
                parameters in `~greykite.algo.forecast.silverkite.forecast_silverkite.SilverkiteForecast.forecast`.

                Allowed keys: ``"silverkite"``, ``"silverkite_diagnostics"``, ``"origin_for_time_vars"``, ``"extra_pred_cols"``,
                ``"fit_algorithm_dict"``, ``"min_admissible_value"``, ``"max_admissible_value"``.

                .. note::

                    ``"extra_pred_cols"`` should contain the desired growth terms, regressor names, and event names.

                ``fit_algorithm_dict`` is a dictionary with ``fit_algorithm`` and ``fit_algorithm_params``
                parameters to `~greykite.algo.forecast.silverkite.forecast_silverkite.SilverkiteForecast.forecast`:

                        fit_algorithm_dict : `dict` or None, optional
                            How to fit the model. A dictionary with the following optional keys.

                            ``"fit_algorithm"`` : `str`, optional, default "linear"
                                The type of predictive model used in fitting.

                                See `~greykite.algo.common.ml_models.fit_model_via_design_matrix`
                                for available options and their parameters.
                            ``"fit_algorithm_params"`` : `dict` or None, optional, default None
                                Parameters passed to the requested fit_algorithm.
                                If None, uses the defaults in `~greykite.algo.common.ml_models.fit_model_via_design_matrix`.

            hyperparameter_override: `dict` [`str`, `any`] or None or `list` [`dict` [`str`, `any`] or None], optional
                After the above model components are used to create a hyperparameter grid, the result is
                updated by this dictionary, to create new keys or override existing ones.
                Allows for complete customization of the grid search.

                Keys should have format ``{named_step}__{parameter_name}`` for the named steps of the
                `sklearn.pipeline.Pipeline` returned by this function. See `sklearn.pipeline.Pipeline`.

                For example::

                    hyperparameter_override={
                        "estimator__origin_for_time_vars": 2018.0,
                        "input__response__null__impute_algorithm": "ts_interpolate",
                        "input__response__null__impute_params": {"orders": [7, 14]},
                        "input__regressors_numeric__normalize__normalize_algorithm": "RobustScaler",
                    }

                If a list of dictionaries, grid search will be done for each dictionary in the list.
                Each dictionary in the list override the defaults. This enables grid search
                over specific combinations of parameters to reduce the search space.

                    * For example, the first dictionary could define combinations of parameters for a
                      "complex" model, and the second dictionary could define combinations of parameters
                      for a "simple" model, to prevent mixed combinations of simple and complex.
                    * Or the first dictionary could grid search over fit algorithm, and the second dictionary
                      could use a single fit algorithm and grid search over seasonality.

                The result is passed as the ``param_distributions`` parameter
                to `sklearn.model_selection.RandomizedSearchCV`.
        model_template: `str`
            This class only accepts "SK".
    """
    DEFAULT_MODEL_TEMPLATE = "SK"
    """The default model template. See `~greykite.framework.templates.model_templates.ModelTemplateEnum`.
    Uses a string to avoid circular imports.
    Overrides the value from `~greykite.framework.templates.forecast_config_defaults.ForecastConfigDefaults`.
    """
    def __init__(
            self,
            estimator: BaseForecastEstimator = SilverkiteEstimator()):
        super().__init__(estimator=estimator)

    @property
    def allow_model_template_list(self):
        """SilverkiteTemplate does not allow `config.model_template` to be a list."""
        return False

    @property
    def allow_model_components_param_list(self):
        """SilverkiteTemplate does not allow `config.model_components_param` to be a list."""
        return False

    def get_regressor_cols(self):
        """Returns regressor column names.

        Implements the method in `~greykite.framework.templates.base_template.BaseTemplate`.

        The intersection of ``extra_pred_cols`` from model components
        and ``self.df`` columns, excluding ``time_col`` and ``value_col``.

        Returns
        -------
        regressor_cols : `list` [`str`] or None
            See `~greykite.framework.pipeline.pipeline.forecast_pipeline`.
        """
        extra_pred_cols = get_extra_pred_cols(model_components=self.config.model_components_param)
        if extra_pred_cols is not None:
            regressor_cols = [col for col in self.df.columns
                              if col not in [self.config.metadata_param.time_col, self.config.metadata_param.value_col]
                              and col in extra_pred_cols]
        else:
            regressor_cols = None
        return regressor_cols

    def get_hyperparameter_grid(self):
        """Returns hyperparameter grid.

        Implements the method in `~greykite.framework.templates.base_template.BaseTemplate`.

        Uses ``self.time_properties`` and ``self.config`` to generate the hyperparameter grid.

        Converts model components and time properties into
        :class:`~greykite.sklearn.estimator.silverkite_estimator.SilverkiteEstimator`
        hyperparameters.

        Notes
        -----
        :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`
        handles the train/test splits according to ``EvaluationPeriodParam``,
        so ``estimator__train_test_thresh`` and ``estimator__training_fraction`` are always None.

        ``estimator__changepoint_detector`` is always None, to prevent leaking future information
        into the past. Pass ``changepoints_dict`` with method="auto" for automatic detection.

        Returns
        -------
        hyperparameter_grid : `dict`, `list` [`dict`] or None
            See :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`.
            The output dictionary values are lists, combined in grid search.
        """
        self.config.model_components_param = apply_default_model_components(
            model_components=self.config.model_components_param,
            time_properties=self.time_properties)

        # returns a single set of parameters for grid search
        hyperparameter_grid = {
            "estimator__silverkite": self.config.model_components_param.custom["silverkite"],
            "estimator__silverkite_diagnostics": self.config.model_components_param.custom["silverkite_diagnostics"],
            "estimator__origin_for_time_vars": self.config.model_components_param.custom["origin_for_time_vars"],
            "estimator__extra_pred_cols": self.config.model_components_param.custom["extra_pred_cols"],
            "estimator__train_test_thresh": [None],
            "estimator__training_fraction": [None],
            "estimator__fit_algorithm_dict": self.config.model_components_param.custom["fit_algorithm_dict"],
            "estimator__daily_event_df_dict": self.config.model_components_param.events["daily_event_df_dict"],
            "estimator__fs_components_df": self.config.model_components_param.seasonality["fs_components_df"],
            "estimator__autoreg_dict": self.config.model_components_param.autoregression["autoreg_dict"],
            "estimator__changepoints_dict": self.config.model_components_param.changepoints["changepoints_dict"],
            "estimator__seasonality_changepoints_dict": self.config.model_components_param.changepoints["seasonality_changepoints_dict"],
            "estimator__changepoint_detector": [None],
            "estimator__min_admissible_value": self.config.model_components_param.custom["min_admissible_value"],
            "estimator__max_admissible_value": self.config.model_components_param.custom["max_admissible_value"],
            "estimator__uncertainty_dict": self.config.model_components_param.uncertainty["uncertainty_dict"],
        }

        # Overwrites values by `model_components.hyperparameter_override`
        # This may produce a list of dictionaries for grid search.
        hyperparameter_grid = update_dictionaries(
            hyperparameter_grid,
            overwrite_dicts=self.config.model_components_param.hyperparameter_override)

        # Ensures all items have the proper type for
        # `sklearn.model_selection.RandomizedSearchCV`.
        # List-type hyperparameters are specified below
        # with their accepted non-list type values.
        hyperparameter_grid = dictionaries_values_to_lists(
            hyperparameter_grid,
            hyperparameters_list_type={
                "estimator__extra_pred_cols": [None]}
        )
        return hyperparameter_grid

    def apply_template_decorator(func):
        """Decorator for ``apply_template_for_pipeline_params`` function.

        Overrides the method in `~greykite.framework.templates.base_template.BaseTemplate`.

        Raises
        ------
        ValueError if config.model_template != "SK"
        """
        @functools.wraps(func)
        def process_wrapper(self, df: pd.DataFrame, config: Optional[ForecastConfig] = None):
            # sets defaults
            config = self.apply_forecast_config_defaults(config)
            # input validation
            if config.model_template != "SK":
                raise ValueError(f"SilverkiteTemplate only supports config.model_template='SK', "
                                 f"found '{config.model_template}'")
            pipeline_params = func(self, df, config)
            return pipeline_params
        return process_wrapper

    @apply_template_decorator
    def apply_template_for_pipeline_params(
            self,
            df: pd.DataFrame,
            config: Optional[ForecastConfig] = None) -> Dict:
        """Explicitly calls the method in
        `~greykite.framework.templates.base_template.BaseTemplate`
        to make use of the decorator in this class.

        Parameters
        ----------
        df : `pandas.DataFrame`
            The time series dataframe with ``time_col`` and ``value_col`` and optional regressor columns.
        config : `~greykite.framework.templates.autogen.forecast_config.ForecastConfig`.
            The `ForecastConfig` class that includes model training parameters.

        Returns
        -------
        pipeline_parameters : `dict`
            The pipeline parameters consumable by
            `~greykite.framework.pipeline.pipeline.forecast_pipeline`.
        """
        return super().apply_template_for_pipeline_params(df=df, config=config)

    apply_template_decorator = staticmethod(apply_template_decorator)
